from google.protobuf import text_format
from ppq.core import NetworkFramework, is_file_exist
from ppq.IR import BaseGraph, GraphBuilder
from ppq.log import NaiveLogger

from .caffe import ppl_caffe_pb2
from .caffe.caffe_graph_optim import de_inplace, merge_batchnorm_scale
from .caffe.caffe_import_utils import caffe_import_map, get_input_shape

logger = NaiveLogger.get_logger('PPQ')

class CaffeParser(GraphBuilder):
    def load_graph_and_format(self, prototxt_path: str, caffemodel_path: str) -> ppl_caffe_pb2.NetParameter:
        if not is_file_exist(prototxt_path):
            raise FileNotFoundError(f'file {prototxt_path} not exist, please check your file path')
        elif not is_file_exist(caffemodel_path):
            raise FileNotFoundError(f'file {caffemodel_path} not existm please check your file path')
        network = ppl_caffe_pb2.NetParameter()
        with open(prototxt_path) as f:
            text_format.Merge(f.read(), network)
        weight = ppl_caffe_pb2.NetParameter()
        with open(caffemodel_path, 'rb') as f:
            weight.ParseFromString(f.read())

        network = de_inplace(network)

        for i in network.layer:
            for j in weight.layer:
                if i.name == j.name:
                    i.ClearField('blobs')
                    i.blobs.MergeFrom(j.blobs)
                    break

        network = merge_batchnorm_scale(network)
        return network

    def build(self, prototxt_path: str, caffemodel_path: str) -> BaseGraph:
        network = self.load_graph_and_format(prototxt_path, caffemodel_path)
        graph = BaseGraph(name=network.name, built_from=NetworkFramework.CAFFE)
        input_shape = get_input_shape(network)
        input_names = list(input_shape.keys())

        activation_shape = input_shape
        top_name_set = set()
        for layer in network.layer:
            if layer.type not in caffe_import_map:
                logger.error(f'{layer.type} Caffe OP is not supported in PPQ import parser yet')
                raise NotImplementedError(f'{layer.type} Caffe OP is not supported in PPQ import parser yet')
            input_shape = [activation_shape[k] for k in layer.bottom]
            caffe_layer = caffe_import_map[layer.type](graph, layer, input_shape)
            graph = caffe_layer.trans()
            activation_shape.update([(k, v) for k, v in zip(layer.top, caffe_layer.out_shape)])

            # statistic top_name and get final out var name
            for name in layer.bottom:
                if name in top_name_set:
                    top_name_set.remove(name)
            for name in layer.top:
                top_name_set.add(name)

        # add input and output for graph
        try:
            for var_name in input_names:
                if var_name not in graph.variables: continue
                graph.inputs[var_name] = graph.variables[var_name]
            for var_name in top_name_set:
                graph.outputs[var_name] = graph.variables[var_name]
        except KeyError as e:
            raise KeyError(
                'seems you got an input/output variable that is not linked to any operation.')
        return graph



"""
## PPQ Foundation Library - PFL

PPQ 基础类库

PFL is a collection of basic classes and functions that provides fundamental functionalities.

    - Parser: Get a network parser.
    
    - Exporter: According to given platform, get a network exporter.
    
    - OperationForwardFunction: According to given platform and optype, get a forward function.
    
    - Dispatcher: Get a network dispatcher.
    
    - FloatingQuantizationConfig: Get a TensorQuantizationConfig for FP8 Quantization.
    
    - LinearQuantizationConfig: Get a TensorQuantizationConfig for INT8 Quantization.
    
    - QuantStub: Get a QuantStub class instance.
    
    - Quantizer: Get a Quantizer corresponding to given platform.
    
    - Observer: Get a Tensor Observer, which is bound to given TensorQuantizationConfig.
    
    - Pipeline: Build Optimization Pipeline.
    
    - QuantFunction: Get PPQ Default Quantize Function.
    
PFL also provides a set of functions to register Quantizer, Parser, Exporter to PPQ.

    - register_network_quantizer
    
    - register_network_parser
    
    - register_network_exporter
    
    - register_operation_handler
    
    - register_calibration_observer
    
"""

from .extension import (register_calibration_observer,
                        register_network_exporter, register_network_parser,
                        register_network_quantizer, register_operation_handler)
from .quant import (Dispatcher, Exporter, FloatingQuantizationConfig,
                    LinearQuantizationConfig, Observer,
                    OperationForwardFunction, Parser, Pipeline, QuantFunction,
                    Quantizer, TensorQuant, ParameterQuant)



# -*- coding: utf-8 -*-
import os
from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Tuple,
    Union,
)

import numpy as np
import onnx
import torch
from onnxsim import simplify
from torch.utils.data import DataLoader

import ppq.lib as PFL
from ppq.api.interface import load_onnx_graph, quantize_onnx_model
from ppq.api.setting import QuantizationSetting, QuantizationSettingFactory
from ppq.core import QuantizationVisibility, TargetPlatform, empty_ppq_cache
from ppq.executor import BaseGraphExecutor, TorchExecutor
from ppq.IR import BaseGraph
from ppq.log import NaiveLogger
from ppq.quantization.analyse import graphwise_error_analyse
from ppq.quantization.analyse.layerwise import layerwise_error_analyse
from ppq.quantization.optim import *

logger = NaiveLogger.get_logger('ESPDL')

def get_target_platform(target: str, num_of_bits: int = 8, float: bool = False):
    platform = None
    if float:
        platform = TargetPlatform.FP32
    else:
        if num_of_bits == 8 and target == "esp32p4":
            platform = TargetPlatform.ESPDL_INT8
        elif num_of_bits == 16 and target == "esp32p4":
            platform = TargetPlatform.ESPDL_INT16
        elif num_of_bits == 8 and target == "esp32s3":
            platform = TargetPlatform.ESPDL_S3_INT8
        elif num_of_bits == 16 and target == "esp32s3":
            platform = TargetPlatform.ESPDL_S3_INT16
        else:
            platform = TargetPlatform.FP32
            logger.warning(f"Do not support num_of_bits:{num_of_bits}, will change to TargetPlatform.FP32")
    
    return platform

def get_random_inputs(input_shape: List[Any], dtype=torch.float32, device='cpu') -> List[Any]:
    if not isinstance(input_shape[0], list):
        input_shape = [input_shape]

    inputs = [
        torch.randn(size=shape, device=device, dtype=dtype)
        for shape in input_shape
    ]
    
    return inputs


def generate_test_value(
    graph: BaseGraph,
    running_device: str,
    inputs: Union[dict, list, torch.Tensor],
    output_names: List[str] = None,
) -> Dict[str, Dict[str, np.ndarray]]:
    test_inputs_value = {}
    test_outputs_value = {}

    executor = TorchExecutor(graph=graph, device=running_device)
    outputs = executor.forward(inputs=inputs, output_names=output_names)
    # get test_inputs_value
    if isinstance(inputs, dict):
        for name, value in inputs.items():
            if name in graph.inputs:
                test_inputs_value[name] = value.clone().detach().cpu()
            else:
                logger.error(f"Can not find input {name} in your graph inputs, please check.")
    else:
        inputs_tmp = executor.prepare_input(inputs=inputs)
        test_inputs_value = {
            name: value.clone().detach().cpu() for name, value in inputs_tmp.items()
        }

    # get test_outputs_value
    if output_names is None:
        outputs_dictionary = graph.outputs
        test_outputs_value = {
            key: outputs[idx].clone().detach().cpu()
            for idx, key in enumerate(outputs_dictionary)
        }
    else:
        test_outputs_value = {
            output_name: output.clone().detach().cpu()
            for output_name, output in zip(output_names, outputs)
        }

    return {"inputs": test_inputs_value, "outputs": test_outputs_value}

def collate_fn_template(batch: Union[torch.Tensor, List[torch.Tensor]], dtype=torch.float32, device='cpu'):
    if isinstance(batch, list) and isinstance(batch[0], torch.Tensor):
        return [x.type(dtype).to(device) for x in batch]
    elif isinstance(batch, torch.Tensor):
        return batch.type(dtype).to(device)
    else:
        logger.error("please provide a valid collate_fn.")

@empty_ppq_cache
def espdl_quantize_onnx(
    onnx_import_file: str,
    espdl_export_file: str,
    calib_dataloader: DataLoader,
    calib_steps: int,
    input_shape: List[Any],
    inputs: List[Any] = None,
    target:str = "esp32p4",
    num_of_bits:int = 8,
    collate_fn: Callable = None,
    dispatching_override: Dict[str, TargetPlatform] = None,
    dispatching_method: str = "conservative",
    setting: QuantizationSetting = None,
    device: str = "cpu",
    error_report: bool = True,
    skip_export: bool = False,
    export_config: bool = True,
    export_test_values: bool = False,
    test_output_names: List[str] = None,
    verbose: int = 0,
) -> BaseGraph :
    """Quantize onnx model and return quantized ppq graph and executor .
    
    Args:
        onnx_import_file (str): onnx model file path
        calib_dataloader (DataLoader): calibration data loader
        calib_steps (int): calibration steps
        input_shape (List[int]):a list of ints indicating size of inputs and batch size must be 1
        inputs (List[str]): a list of Tensor and batch size must be 1
        target: target chip, support "esp32p4" and "esp32s3"
        num_of_bits: the number of quantizer bits, 8 or 16
        collate_fn (Callable): batch collate func for preprocessing
        dispatching_override (deprecated): override dispatching result.
                                           It is recommended to use the `setting` parameter, as this argument will be removed in the next version.
        dispatching_method (deprecated): Refer to https://github.com/espressif/esp-ppq/blob/master/ppq/scheduler/__init__.py#L8.
                                         It is recommended to use the `setting` parameter, as this argument will be removed in the next version.
        setting (QuantizationSetting): Quantization setting, default espdl setting will be used when set None
        device (str, optional):  execution device, defaults to 'cpu'.
        error_report (bool, optional): whether to print error report, defaults to True.
        skip_export (bool, optional): whether to export the quantized model, defaults to False.
        export_config (bool, optional): whether to export the quantization configuration, defaults to True.
        export_test_values (bool, optional): whether to export the test values, defaults to False.
        test_output_names (List[str], optional): tensor names of the model want to test, defaults to None.
        verbose (int, optional): whether to print details, defaults to 0.

    Returns:
        BaseGraph:      The Quantized Graph, containing all information needed for backend execution
    """    
    
    export_path = os.path.dirname(os.path.abspath(espdl_export_file))
    os.makedirs(export_path, exist_ok=True)

    # ------------------------------------------------------------
    #
    # 1: Quantize ONNX Model.
    #
    #  ------------------------------------------------------------
    if calib_dataloader is None or calib_steps is None:
        raise TypeError(
            "Quantization needs a valid calib_dataloader and calib_steps setting."
        )
    target_platform = get_target_platform(target, num_of_bits)
    input_dtype = torch.float32
    if num_of_bits == 16:
        input_dtype = torch.float64
    
    if not collate_fn:
        collate_fn = partial(collate_fn_template, dtype=input_dtype, device=device)

    ppq_graph = None
    if inputs:
        dummy_inputs = inputs
    else:
        dummy_inputs = get_random_inputs(input_shape, input_dtype, device)

    if target_platform != TargetPlatform.FP32:
        if dispatching_override is not None or dispatching_method != "conservative":
            logger.warning(f"It is recommended to use the setting parameter. The dispatching_override and dispatching_method will be deprecated.")

        if setting is None:
            setting = QuantizationSettingFactory.espdl_setting()

        if dispatching_method != "conservative":
            setting.dispatcher = dispatching_method

        if dispatching_override is not None:
            for opname, platform in dispatching_override.items():
                setting.dispatching_table.append(opname, platform)

        ppq_graph = quantize_onnx_model(
                        onnx_import_file=onnx_import_file,
                        calib_dataloader=calib_dataloader,
                        calib_steps=calib_steps,
                        input_shape=None,
                        platform=target_platform,
                        input_dtype=input_dtype,
                        inputs=dummy_inputs,
                        setting=setting,
                        collate_fn=collate_fn,
                        device=device,
                        verbose=verbose,
                        do_quantize=True,
                    )

        # ------------------------------------------------------------
        #
        # 2: Analyze Quantization Errors.
        #
        # ------------------------------------------------------------
        if error_report:
            graphwise_error_analyse(
                graph=ppq_graph,
                running_device=device,
                collate_fn=collate_fn,
                dataloader=calib_dataloader,
            )

            layerwise_error_analyse(
                graph=ppq_graph,
                running_device=device,
                collate_fn=collate_fn,
                dataloader=calib_dataloader,
            )
    else:
        # support TargetPlatform.FP32
        ppq_graph = load_onnx_graph(onnx_import_file=onnx_import_file)
        executor = TorchExecutor(graph=ppq_graph, device=device)
        executor.tracing_operation_meta(inputs=dummy_inputs)
        target_platform = TargetPlatform.ESPDL_INT8
    
    # ------------------------------------------------------------
    #
    # 3: Export ESPDL Model.
    #
    # ------------------------------------------------------------
    if not skip_export:

        values_for_test = None
        if export_test_values:
            values_for_test = generate_test_value(ppq_graph, device, dummy_inputs, test_output_names)

        PFL.Exporter(platform=target_platform).export(
            file_path=espdl_export_file,
            graph=ppq_graph,
            values_for_test=values_for_test,
            export_config=export_config,
        )
    return ppq_graph


def espdl_quantize_torch(
    model: torch.nn.Module,
    espdl_export_file: str,
    calib_dataloader: DataLoader,
    calib_steps: int,
    input_shape: List[Any],
    inputs: List[Any] = None,
    target:str = "esp32p4",
    num_of_bits:int = 8,
    collate_fn: Callable = None,
    dispatching_override: Dict[str, TargetPlatform] = None,
    dispatching_method: str = "conservative",
    setting: QuantizationSetting = None,
    device: str = "cpu",
    error_report: bool = True,
    skip_export: bool = False,
    export_config: bool = True,
    export_test_values: bool = False,
    test_output_names: List[str] = None,
    verbose: int = 0,
) -> BaseGraph:
    """Quantize torch model and return quantized ppq graph and executor .
    
    Args:
        model (torch.nn.Module): torch model
        calib_dataloader (DataLoader): calibration data loader
        calib_steps (int): calibration steps
        input_shape (List[int]):a list of ints indicating size of inputs and batch size must be 1
        inputs (List[str]): a list of Tensor and batch size must be 1
        target: target chip, support "esp32p4" and "esp32s3"
        num_of_bits: the number of quantizer bits, 8 or 16
        collate_fn (Callable): batch collate func for preprocessing
        dispatching_override (deprecated): override dispatching result.
                                        It is recommended to use the `setting` parameter, as this argument will be removed in the next version.
        dispatching_method (deprecated): Refer to https://github.com/espressif/esp-ppq/blob/master/ppq/scheduler/__init__.py#L8.
                                        It is recommended to use the `setting` parameter, as this argument will be removed in the next version.
        setting (QuantizationSetting): Quantization setting, default espdl setting will be used when set None
        device (str, optional):  execution device, defaults to 'cpu'.
        error_report (bool, optional): whether to print error report, defaults to True.
        skip_export (bool, optional): whether to export the quantized model, defaults to False.
        export_config (bool, optional): whether to export the quantization configuration, defaults to True.
        export_test_values (bool, optional): whether to export the test values, defaults to False.
        test_output_names (List[str], optional): tensor names of the model want to test, defaults to None.
        verbose (int, optional): whether to print details, defaults to 0.

    Returns:
        BaseGraph:      The Quantized Graph, containing all information needed for backend execution
    """   
    if not isinstance(input_shape[0], list):
        input_shape = [input_shape]
    export_path = os.path.dirname(os.path.abspath(espdl_export_file))
    os.makedirs(export_path, exist_ok=True)

    # step1: export onnx model    
    model = model.eval()
    model = model.to(device)
    if num_of_bits == 16:
        model = model.double()

    base_file_name, _ = os.path.splitext(espdl_export_file)
    onnx_file_path = base_file_name + ".onnx"
    torch.onnx.export(
        model=model,
        args=tuple(
            [
                torch.zeros(
                    size=shape,
                    device=device,
                    dtype=torch.float32 if num_of_bits == 8 else torch.float64,
                )
                for shape in input_shape
            ]
        ),
        f=onnx_file_path,
        opset_version=13,
        do_constant_folding=True,
    )

    model = onnx.load(onnx_file_path)
    model_sim, check = simplify(model)
    if check:
        onnx.save(model_sim, onnx_file_path)
    
    # step2: quantize onnx model and export espdl model
    return espdl_quantize_onnx(
        onnx_import_file=onnx_file_path,
        espdl_export_file=espdl_export_file,
        calib_dataloader=calib_dataloader,
        calib_steps=calib_steps,
        input_shape=input_shape,
        inputs=inputs,
        target=target,
        num_of_bits=num_of_bits,
        collate_fn=collate_fn,
        dispatching_override=dispatching_override,
        dispatching_method=dispatching_method,
        setting=setting,
        device=device,
        error_report=error_report,
        skip_export=skip_export,
        export_config=export_config,
        export_test_values=export_test_values,
        test_output_names=test_output_names,
        verbose=verbose,
    )



from typing import Union

import torch
from ppq.api.setting import QuantizationSetting
from ppq.core import (PASSIVE_OPERATIONS, OperationQuantizationConfig,
                      QuantizationPolicy, QuantizationProperty,
                      QuantizationStates, RoundingPolicy, TargetPlatform)
from ppq.IR import BaseGraph, Operation
from ppq.quantization.optim import (NxpInputRoundingRefinePass,
                                    NXPResizeModeChangePass,
                                    QuantizationOptimizationPipeline)

from .base import BaseQuantizer


class NXP_Quantizer(BaseQuantizer):
    def __init__(
        self,
        graph: BaseGraph,
    ) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = - int(pow(2, self._num_of_bits - 1))
        self._quant_max = int(pow(2, self._num_of_bits - 1) - 1)

    def build_quant_pipeline(self, setting: QuantizationSetting) -> QuantizationOptimizationPipeline:
        pipeline = super().build_quant_pipeline(setting)
        pipeline.append_optimization_to_pipeline(NXPResizeModeChangePass(), at_front=True)
        pipeline.append_optimization_to_pipeline(NxpInputRoundingRefinePass(), at_front=True)
        return pipeline

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:

        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy, rounding=self.rounding_policy,
            op=operation, num_of_bits=self._num_of_bits, exponent_bits=0,
            quant_max=self._quant_max, quant_min=self._quant_min,
            observer_algorithm='percentile')

        if operation.type in {'Conv', 'Gemm'}:
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel quant-config.
            assert operation.num_of_input > 0, 'Seems you got a Conv layer with no parameters.'

            # first parameter must exits, for conv layer it will be conv_weight
            # layout: [out_channel, in_channel, kernel_size, kernel_size]
            if operation.type in {'Conv'}:
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.LINEAR +
                    QuantizationProperty.PER_CHANNEL +
                    QuantizationProperty.POWER_OF_2
                )
                conv_weight_config.rounding = RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO
                conv_weight_config.channel_axis = 0
                conv_weight_config.observer_algorithm = 'minmax'
            # first parameter must exits, for gemm layer it will be gemm_weight
            # layout: [in_dim, out_dim]
            elif operation.type in {'Gemm'}:
                gemm_weight_config = base_quant_config.input_quantization_config[1]
                gemm_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.LINEAR +
                    QuantizationProperty.PER_CHANNEL +
                    QuantizationProperty.POWER_OF_2
                )
                gemm_weight_config.rounding = RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO
                gemm_weight_config.channel_axis = 0
                gemm_weight_config.observer_algorithm = 'minmax'

            # if operation has bias
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.LINEAR +
                    QuantizationProperty.PER_CHANNEL +
                    QuantizationProperty.POWER_OF_2
                )
                bias_config.num_of_bits = 30
                bias_config.quant_max = int(pow(2, 30))
                bias_config.quant_min = - int(pow(2, 30))
                bias_config.state = QuantizationStates.PASSIVE_INIT
                bias_config.channel_axis = 0
                bias_config.observer_algorithm = 'minmax'

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False

        return base_quant_config

    @ property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.NXP_INT8

    @ property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @ property
    def quant_operation_types(self) -> set:
        return {
            'Conv', 'Gemm', 'Relu', 'PRelu',
            'Clip', 'Pad', 'Resize', 'MaxPool', 'AveragePool',
            'GlobalMaxPool', 'GlobalAveragePool',
            'Mul', 'Add', 'Max', 'Sub', 'Div',
            'LeakyRelu', 'Concat', 'Sigmoid', 'Slice'
        }

    @ property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL +
            QuantizationProperty.LINEAR +
            QuantizationProperty.PER_TENSOR +
            QuantizationProperty.POWER_OF_2
        )

    @ property
    def rounding_policy(self):
        return RoundingPolicy.ROUND_HALF_UP

    @ property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip'}



from tmodel  import *
from tscheme import *
from ppq     import *
from ppq.api import *
import sys

DEVICE = 'cuda'

with ENABLE_CUDA_KERNEL():
    for scheme in TEST_SCHEMES:
        for case in TORCH_TEST_CASES:
            try:
                print(f'PPQ System test start with model {case.model_name}, Scheme: {scheme.name}')
                dataset = [case.input_generator().to(DEVICE) for _ in range(8)]
                model = case.model_builder().to(DEVICE)

                quantized = quantize_torch_model(
                    model=model,
                    calib_dataloader=dataset,
                    calib_steps=8,
                    input_shape=case.input_generator().shape,
                    platform=scheme.quant_platform,
                    setting=scheme.setting)

                if (case.deploy_platforms is None or
                    scheme.export_platform in case.deploy_platforms):
                    export_ppq_graph(
                        graph=quantized,
                        platform=scheme.export_platform,
                        graph_save_to='tworkingspace/export',
                        config_save_to='tworkingspace/export.json')
            except NotImplementedError as e:
                print(f'{time.strftime("%Y-%m-%d %H:%M:%S")} | Error occurred: {e}')
                sys.exit(1)



from pickle import dump, load

from ppq.core import PPQ_CONFIG
from ppq.IR import BaseGraph, GraphExporter
from ppq.IR.base.graph import GraphBuilder


class NativeExporter(GraphExporter):
    def __init__(self) -> None:
        super().__init__()
    def export(self, file_path: str, graph: BaseGraph,
               config_path: str = None, dump_value: bool = True):
        def dump_elements_to_file(file, elements: list):
            for element in elements: dump(element, file)

        with open(file_path, 'wb') as file:
            dump_elements_to_file(file, elements=[
                'PPQ GRAPH DEFINITION', # PPQ Signature.
                PPQ_CONFIG.VERSION,     # PPQ Signature.
                graph,
            ])

class NativeImporter(GraphBuilder):
    def __init__(self) -> None:
        super().__init__()

    def build(self, file_path: str, **kwargs) -> BaseGraph:
        def load_elements_from_file(file, num_of_elements: int) -> list:
            try: return [load(file) for _ in range(num_of_elements)]
            except EOFError as e:
                raise Exception('File format parsing error. Unexpected EOF found.')

        with open(file_path, 'rb') as file:
            signature, version, graph = load_elements_from_file(file, 3)
            if signature != 'PPQ GRAPH DEFINITION':
                raise Exception('File format parsing error. Graph Signature has been damaged.')
            if str(version) > PPQ_CONFIG.VERSION:
                print(f'\033[31mWarning: Dump file is created by PPQ({str(version)}), '
                f'however you are using PPQ({PPQ_CONFIG.VERSION}).\033[0m')

        assert isinstance(graph, BaseGraph), (
            'File format parsing error. Graph Definition has been damaged.')
        try:
            for op in graph.operations.values():
                input_copy, _ = op.inputs.copy(), op.inputs.clear()
                for name in input_copy: op.inputs.append(graph.variables[name])
                output_copy, _ = op.outputs.copy(), op.outputs.clear()
                for name in output_copy: op.outputs.append(graph.variables[name])

            for var in graph.variables.values():
                dest_copy, _ = var.dest_ops.copy(), var.dest_ops.clear()
                for name in dest_copy: var.dest_ops.append(graph.operations[name])
                if var.source_op is not None:
                    var.source_op = graph.operations[var.source_op]

            graph._graph_inputs = {name: graph.variables[name] for name in graph._graph_inputs}
            graph._graph_outputs = {name: graph.variables[name] for name in graph._graph_outputs}
        except Exception as e:
            raise Exception('File format parsing error. Graph Definition has been damaged.')
        return graph



# 这里我们展示两种不同的方法去生成 TensorRT Engine

# Plan B: PPQ 导出 engine

import os
import torchvision.transforms as transforms
from PIL import Image
from ppq import *
from ppq.api import *

ONNX_PATH        = 'models/yolov5s6.onnx'      # 你的模型位置
OUTPUT_PATH      = 'Output'                    # 生成的量化模型的位置
CALIBRATION_PATH = 'imgs'                      # 校准数据集
BATCHSIZE        = 1
EXECUTING_DEVICE = 'cuda'
# create dataloader
imgs = []
trans = transforms.Compose([
    transforms.Resize([640, 640]),  # [h,w]
    transforms.ToTensor(),
])
for file in os.listdir(path=CALIBRATION_PATH):
    path = os.path.join(CALIBRATION_PATH, file)
    img = Image.open(path).convert('RGB')
    img = trans(img)
    imgs.append(img) # img is 0 - 1

from torch.utils.data import DataLoader
dataloader = DataLoader(dataset=imgs, batch_size=BATCHSIZE)

with ENABLE_CUDA_KERNEL():
    qir = quantize_onnx_model(
        platform=TargetPlatform.TRT_INT8,
        onnx_import_file=ONNX_PATH, 
        calib_dataloader=dataloader, 
        calib_steps=32, device=EXECUTING_DEVICE,
        input_shape=[BATCHSIZE, 3, 640, 640], 
        collate_fn=lambda x: x.to(EXECUTING_DEVICE))

    snr_report = graphwise_error_analyse(
        graph=qir, running_device=EXECUTING_DEVICE, 
        dataloader=dataloader, collate_fn=lambda x: x.to(EXECUTING_DEVICE))

    snr_report = layerwise_error_analyse(
        graph=qir, running_device=EXECUTING_DEVICE, 
        dataloader=dataloader, collate_fn=lambda x: x.to(EXECUTING_DEVICE))

    export_ppq_graph(
        qir, platform=TargetPlatform.TRT_INT8, 
        graph_save_to=OUTPUT_PATH.join('/INT8.onnx'), 
        config_save_to=OUTPUT_PATH.join('/INT8.json'))
    
    from ppq.utils.TensorRTUtil import build_engine, Benchmark, Profiling
    build_engine(
        onnx_file=OUTPUT_PATH.join('/INT8.onnx'), 
        engine_file=OUTPUT_PATH.join('/INT8.engine'), int8=True, 
        int8_scale_file=OUTPUT_PATH.join('/INT8.json'))

    Benchmark(OUTPUT_PATH.join('/INT8.engine'))
    Profiling(OUTPUT_PATH.join('/INT8.engine'))


from ppq import BaseGraph
from ppq.core.quant import NetworkFramework
from ppq.scheduler.core.opsocket import OType
from ppq.scheduler.perseus import Perseus

graph = BaseGraph(name='TestCase1', built_from=NetworkFramework.ONNX)
gemm = graph.create_operation(op_type='Gemm', name='FirstGemm')
conv = graph.create_operation(op_type='Conv', name='Conv')
# g1.create_operation(op_type='Shape')
sigmoid = graph.create_operation(op_type='Sigmoid', name='Sigmoid')
shape1  = graph.create_operation(op_type='Shape', name='Shape1')
relu    = graph.create_operation(op_type='Relu', name='Relu')
gemm2   = graph.create_operation(op_type='Gemm', name='ShapeGemm')
shape2  = graph.create_operation(op_type='Shape', name='Shape2')
reshape = graph.create_operation(op_type='Reshape', name='FinalReshape')
gemm3   = graph.create_operation(op_type='Gemm', name='FinalGemm')

graph.create_link_with_op(graph.create_variable(), A=None, B=gemm)
graph.create_link_with_op(graph.create_variable(), A=gemm, B=sigmoid)
graph.create_link_with_op(graph.create_variable(), A=sigmoid, B=shape1)
graph.create_link_with_op(graph.create_variable(), A=shape1, B=relu)
graph.create_link_with_op(graph.create_variable(), A=relu, B=gemm2)
graph.create_link_with_op(graph.create_variable(), A=gemm2, B=None)
graph.create_link_with_op(graph.create_variable(), A=conv, B=reshape)
graph.create_link_with_op(graph.create_variable(), A=relu, B=shape2)
graph.create_link_with_op(graph.create_variable(), A=shape2, B=reshape)
graph.create_link_with_op(graph.create_variable(), A=reshape, B=gemm3)
graph.create_link_with_op(graph.create_variable(), A=None, B=conv)
dispatcher = Perseus(graph)

for op in graph.operations.values():
    print(op.name, [op.name for op in dispatcher.parse_transitive_fanout(parsing_from=[op])])
dispatch_table = dispatcher.dispatch()
print(dispatch_table)

assert dispatch_table['FirstGemm'] == OType.QUANTABLE
assert dispatch_table['Shape1'] == OType.NONQUANTABLE
assert dispatch_table['Shape2'] == OType.NONQUANTABLE
assert dispatch_table['ShapeGemm'] == OType.CONTROVERSIAL
assert dispatch_table['FinalReshape'] == OType.QUANTABLE
assert dispatch_table['FinalGemm'] == OType.QUANTABLE

graph = BaseGraph(name='TestCase2', built_from=NetworkFramework.ONNX)
gemm = graph.create_operation(op_type='Gemm', name='FirstGemm')
topk = graph.create_operation(op_type='TopK', name='TopK')
gemm2 = graph.create_operation(op_type='Gemm', name='Gemm2')
gemm3 = graph.create_operation(op_type='Gemm', name='Gemm3')
mul = graph.create_operation(op_type='Mul', name='Mul')

graph.create_link_with_op(graph.create_variable(), A=None, B=gemm)
graph.create_link_with_op(graph.create_variable(), A=gemm, B=topk)
graph.create_link_with_op(graph.create_variable(), A=topk, B=gemm2)
graph.create_link_with_op(graph.create_variable(), A=topk, B=gemm3)
graph.create_link_with_op(graph.create_variable(), A=gemm2, B=mul)
graph.create_link_with_op(graph.create_variable(), A=gemm3, B=mul)
graph.create_link_with_op(graph.create_variable(), A=mul, B=None)

dispatcher = Perseus(graph)
dispatch_table = dispatcher.dispatch()
print(dispatch_table)

for op in graph.operations.values():
    print(op.name, [op.name for op in dispatcher.parse_transitive_fanout(parsing_from=[op])])

assert dispatch_table['FirstGemm'] == OType.CONTROVERSIAL
assert dispatch_table['Gemm2'] == OType.CONTROVERSIAL
assert dispatch_table['Gemm3'] == OType.CONTROVERSIAL
assert dispatch_table['Mul'] == OType.CONTROVERSIAL


graph = BaseGraph(name='TestCase3', built_from=NetworkFramework.ONNX)
gemm = graph.create_operation(op_type='Gemm', name='FirstGemm')
reshape = graph.create_operation(op_type='Reshape', name='Reshape')
add = graph.create_operation(op_type='Add', name='Add')
shape = graph.create_operation(op_type='Shape', name='Shape')
const = graph.create_operation(op_type='Constant', name='Constant')
conv = graph.create_operation(op_type='Conv', name='Conv')
mul = graph.create_operation(op_type='Mul', name='Mul')
cos = graph.create_operation(op_type='ConstantOfShape', name='ConstantOfShape')
shape2 = graph.create_operation(op_type='Shape', name='Shape2')

graph.create_link_with_op(graph.create_variable(), A=None, B=gemm)
graph.create_link_with_op(graph.create_variable(), A=gemm, B=reshape)
graph.create_link_with_op(graph.create_variable(), A=None, B=shape)
graph.create_link_with_op(graph.create_variable(), A=None, B=shape2)
graph.create_link_with_op(graph.create_variable(), A=shape, B=add)
graph.create_link_with_op(graph.create_variable(), A=const, B=add)
graph.create_link_with_op(graph.create_variable(), A=add, B=reshape)

graph.create_link_with_op(graph.create_variable(), A=reshape, B=conv)
graph.create_link_with_op(graph.create_variable(), A=conv, B=mul)
graph.create_link_with_op(graph.create_variable(), A=shape2, B=cos)
graph.create_link_with_op(graph.create_variable(), A=cos, B=mul)
graph.create_link_with_op(graph.create_variable(), A=mul, B=None)

dispatcher = Perseus(graph)
print([op.name for op in dispatcher.parse_transitive_fanout(parsing_from=[conv])])
dispatch_table = dispatcher.dispatch()
print(dispatch_table)

assert dispatch_table['FirstGemm'] == OType.QUANTABLE
assert dispatch_table['Add']       == OType.NONQUANTABLE
assert dispatch_table['Constant']  == OType.NONQUANTABLE
assert dispatch_table['Mul']       == OType.QUANTABLE
assert dispatch_table['Conv']      == OType.QUANTABLE
assert dispatch_table['Reshape']   == OType.QUANTABLE




# This file is created by Nvidia Corp.
# Modified by PPQ develop team.
#
# Copyright 2020 NVIDIA Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import struct
from typing import List

from ppq.core import (NetworkFramework, QuantizationPolicy,
                      QuantizationProperty, ppq_info, ppq_warning)
from ppq.IR import BaseGraph, GraphExporter
from ppq.IR.quantize import QuantableOperation

from .caffe_exporter import CaffeExporter
from .onnxruntime_exporter import OnnxExporter, ONNXRUNTIMExporter


class TensorRTExporter_QDQ(ONNXRUNTIMExporter):
    """
    TensorRT PPQ 0.6.4 以来新加入的功能
    
    你需要注意，只有 TensorRT 8.0 以上的版本支持读取 PPQ 导出的量化模型
    并且 TensorRT 对于量化模型的解析存在一些 Bug，
    
    如果你遇到模型解析不对的问题，欢迎随时联系我们进行解决。
    
        已知的问题包括：
        1. 模型导出时最好不要包含其他 opset，如果模型上面带了别的opset，比如 mmdeploy，trt有可能会解析失败
        2. 模型导出时可能出现 Internal Error 10, Invalid Node xxx()，我们还不知道如何解决该问题
    
    Args:
        ONNXRUNTIMExporter (_type_): _description_
    """

    def export(self, file_path: str, graph: BaseGraph,
               save_as_external_data: bool = False) -> None:
        # step 1, export onnx file.
        super().export(
            file_path=file_path, graph=graph, quantized_param=False,
            config_path=None, save_as_external_data=save_as_external_data)


class TensorRTExporter_JSON(GraphExporter):
    """
    使用该导出器导出符合 TensorRT 格式要求的量化模型。
    
    PPQ 可以将 TQC 中的量化信息导出到 json，并遵循 TensorRT 支持的格式。
    用户可以通过 Json 文件将图中所有 variable 的 min, max 传递给 TensorRT
    所有未写入 min, max 的变量将保持 fp32(fp16) 精度执行。
    """
    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        act_quant_info = {}
        for op in graph.topological_sort():
            if not isinstance(op, QuantableOperation): continue
            if op.type in {"Gather", "Unsqueeze", "Concat", "Reshape", "Squeeze"}: continue

            for cfg, var in op.config_with_variable:
                if not cfg.can_export(export_overlapped=True): continue
                if var.is_parameter: continue

                if cfg.policy != QuantizationPolicy(
                    QuantizationProperty.LINEAR + 
                    QuantizationProperty.SYMMETRICAL + 
                    QuantizationProperty.PER_TENSOR):
                    ppq_warning(f'Can not export quantization config on variable {var.name}, '
                                'Quantization Policy is invalid.')
                    continue

                if cfg.num_of_bits != 8 or cfg.quant_max != 127 or cfg.quant_min != -128:
                    ppq_warning(f'Can not export quantization config on variable {var.name}, '
                                'Tensor Quantization Config has unexpected setting.')
                    continue

                act_quant_info[var.name] = cfg.scale.item() * 127

        json_qparams_str = json.dumps({'act_quant_info': act_quant_info}, indent=4)
        with open(config_path, "w") as json_file:
            json_file.write(json_qparams_str)

    def export_weights(self, graph: BaseGraph, config_path: str = None):
        topo_order =  graph.topological_sort()
        weights_list = []
        for index, op in enumerate(topo_order):
            if op.type in {"Conv", "Gemm"}:
                weights_list.extend(op.parameters)

        weight_file_path = os.path.join(os.path.dirname(config_path), "quantized.wts")

        f = open(weight_file_path, 'w')
        f.write("{}\n".format(len(weights_list)))

        for param in weights_list:
            weight_name = param.name
            weight_value = param.value.reshape(-1).cpu().numpy()
            f.write("{} {}".format(weight_name, len(weight_value)))
            for value in weight_value:
                f.write(" ")
                f.write(struct.pack(">f", float(value)).hex())
            f.write("\n")
        ppq_info(f'Parameters have been saved to file: {weight_file_path}')


    def export(self, file_path: str, graph: BaseGraph, config_path: str = None, input_shapes: List[List[int]] = [[1, 3, 224, 224]]):
        ppq_info(
            'You are exporting PPQ Graph to TensorRT(Onnx + Json). \n'
            'Please Compile the TensorRT INT8 engine manually: \n\n'
            'from ppq.utils.TensorRTUtil import build_engine \n'
            "build_engine(onnx_file='Quantized.onnx', int8_scale_file='Quantized.json', engine_file='Quantized.engine', int8=True)\n")
        if config_path is not None:
            self.export_quantization_config(config_path, graph)
            self.export_weights(graph, config_path)
        _, ext = os.path.splitext(file_path)
        if ext == '.onnx':
            exporter = OnnxExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None)
        elif ext in {'.prototxt', '.caffemodel'}:
            exporter = CaffeExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None, input_shapes=input_shapes)

        # no pre-determined export format, we export according to the
        # original model format
        elif graph._built_from == NetworkFramework.CAFFE:
            exporter = CaffeExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None, input_shapes=input_shapes)
        elif graph._built_from == NetworkFramework.ONNX:
            exporter = OnnxExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None)



from ppq.core import CUDA
from ppq.core.ffi import CUDA_COMPLIER
from ppq import ppq_tensor_round, RoundingPolicy, ppq_numerical_round, torch_snr_error
from typing import List
import torch
from math import sqrt
from tqdm import tqdm

EXECUTING_DEVICE = 'cuda'
TEST_TIMES = 128
Q_MIN, Q_MAX = 0, 255
ROUNDING_POLICY = RoundingPolicy.ROUND_HALF_EVEN

CUDA_COMPLIER.complie()


def __TEST_QUANTIZE_LT__(size: List[int], iterations: int, sym: bool):
    for _ in tqdm(range(iterations), desc='QUANTIZE LT TESTING...'):
        t = torch.rand(size=size).cuda() * 32
        s = torch.rand(size=[1]).cuda()
        if sym:
            o = torch.zeros(size=[1]).cuda()
        else:
            o = torch.randint(low=0, high=255, size=[1]).float().cuda()

        # torch quantize
        qt = ppq_tensor_round(t / s, policy=ROUNDING_POLICY) + o
        qt = qt.clip(Q_MIN, Q_MAX)
        qt = (qt - o) * s

        # t = t.int()
        cuda_qt = CUDA.LinearQuantize_T(
            t, s, o, Q_MIN, Q_MAX, ROUNDING_POLICY.value)

        diff = qt - cuda_qt
        if diff.abs().max() != 0:
            raise Exception('Test Failed.')


def __TEST_QUANTIZE_LC__(size: List[int], iterations: int, sym: bool, c: int=1):
    for _ in tqdm(range(iterations), desc='QUANTIZE LC TESTING...'):
        num_of_channel = size[c]
        t = torch.rand(size=size).cuda() * 32
        s = torch.rand(size=[num_of_channel]).cuda()
        if sym:
            o = torch.zeros(size=[num_of_channel]).cuda()
        else:
            o = torch.randint(low=0, high=255, size=[num_of_channel]).float().cuda()

        shape = [1 if axis != c else -1 for axis in range(t.ndim)]
        s, o = s.view(shape), o.view(shape)
        qt = ppq_tensor_round(t / s, policy=ROUNDING_POLICY) + o
        qt = qt.clip(Q_MIN, Q_MAX)
        qt = (qt - o) * s

        # t = t.int()
        cuda_qt = CUDA.LinearQuantize_C(
            t, s, o, c, Q_MIN, Q_MAX, ROUNDING_POLICY.value)

        diff = qt - cuda_qt
        if diff.abs().max() != 0:
            print(diff)
            raise Exception('Test Failed.')


def __TEST_QUANTIZE_LT_B__(size: List[int], iterations: int, sym: bool):
    def ref_grad_func(value: torch.Tensor, dy: torch.Tensor, 
                      scale: torch.Tensor, offset: torch.Tensor):
        
        qt = ppq_tensor_round(value / scale, policy=ROUNDING_POLICY) + offset
        clipped_qt = qt.clip(Q_MIN, Q_MAX)
        
        dx = torch.where(clipped_qt != qt, torch.zeros_like(dy), dy)
        ds = torch.where(clipped_qt == qt, (((qt - offset) * scale) - value) * dy / scale, torch.zeros_like(dy))
        ds += torch.where(qt > Q_MAX, (Q_MAX - offset) * dy, torch.zeros_like(dy))
        ds += torch.where(qt < Q_MIN, (Q_MIN - offset) * dy, torch.zeros_like(dy))
        ds = ds.sum() / sqrt(value.numel() * (Q_MAX - Q_MIN))
        return dx, ds

    for _ in tqdm(range(iterations), desc='QUANTIZE LT_B TESTING...'):
        t = torch.rand(size=size).cuda() * 50
        s = torch.rand(size=[1]).cuda()
        if sym:
            o = torch.zeros(size=[1]).cuda()
        else:
            o = torch.randint(low=0, high=255, size=[1]).float().cuda()
        dy = torch.rand_like(t)
        
        cuda_grad_x, cuda_grad_s = CUDA.LinearQuantize_T_B(
            t, s, o, dy, Q_MIN, Q_MAX, ROUNDING_POLICY.value)
        
        ref_grad_x, ref_grad_s = ref_grad_func(t, dy, s, o)
        diff = ref_grad_x.flatten() - cuda_grad_x.flatten()
        if diff.abs().max() != 0:
            raise Exception('Test Failed.')

        snr = torch_snr_error(cuda_grad_s.reshape([1, -1]), ref_grad_s.reshape([1, -1])).item()
        if snr > 1e-3:
            print(f'[Warning]. Floating Precision Error. {snr:.4f} {cuda_grad_s.item():4f} {ref_grad_s.item():4f}')
            if cuda_grad_s.item() > 1:
                raise Exception('Test Failed.')
        
def __TEST_QUANTIZE_LC_B__(size: List[int], iterations: int, sym: bool, c: int=1):
    def ref_grad_func(value: torch.Tensor, dy: torch.Tensor, 
                      scale: torch.Tensor, offset: torch.Tensor, c: int):

        shape = [1 if axis != c else -1 for axis in range(t.ndim)]
        scale = scale.view(shape)
        offset = offset.view(shape)
        
        qt = ppq_tensor_round(value / scale, policy=ROUNDING_POLICY) + offset
        clipped_qt = qt.clip(Q_MIN, Q_MAX)

        dx = torch.where(clipped_qt != qt, torch.zeros_like(dy), dy)
        ds = torch.where(clipped_qt == qt, (((qt - offset) * scale) - value) * dy / scale, torch.zeros_like(dy))
        ds += torch.where(qt > Q_MAX, (Q_MAX - offset) * dy, torch.zeros_like(dy))
        ds += torch.where(qt < Q_MIN, (Q_MIN - offset) * dy, torch.zeros_like(dy))
        ds = ds.transpose(0, c).flatten(1).sum(dim=-1) / sqrt(value.numel() * (Q_MAX - Q_MIN))
        return dx, ds

    for _ in tqdm(range(iterations), desc='QUANTIZE LC_B TESTING...'):
        num_of_channel = size[c]
        t = torch.rand(size=size).cuda() * 32
        s = torch.rand(size=[num_of_channel]).cuda()
        if sym:o = torch.zeros(size=[num_of_channel]).cuda()
        else: o = torch.randint(low=0, high=255, size=[num_of_channel]).float().cuda()
        dy = torch.rand_like(t)
        
        cuda_grad_x, cuda_grad_s = CUDA.LinearQuantize_C_B(
            t, s, o, dy, Q_MIN, Q_MAX, c, ROUNDING_POLICY.value)
        
        ref_grad_x, ref_grad_s = ref_grad_func(t, dy, s, o, c)
        diff = ref_grad_x.flatten() - cuda_grad_x.flatten()
        if diff.abs().max() != 0:
            print(ref_grad_x)
            print(cuda_grad_x)
            raise Exception('Test Failed.')

        snr = torch_snr_error(cuda_grad_s.reshape([1, -1]), ref_grad_s.reshape([1, -1])).item()
        if snr > 1e-5:
            print(f'[Warning]. Floating Precision Error. {snr:.4f} {cuda_grad_s.item():4f} {ref_grad_s.item():4f}')
            if cuda_grad_s.item() > 1:
                raise Exception('Test Failed.')

__TEST_QUANTIZE_LT__(size=[1, 1, 1, 1], iterations=100, sym=True)
__TEST_QUANTIZE_LT__(size=[1, 1, 1, 1], iterations=100, sym=False)
__TEST_QUANTIZE_LT__(size=[5, 12, 13, 4], iterations=100, sym=True)
__TEST_QUANTIZE_LT__(size=[1, 7, 15, 41], iterations=100, sym=False)
__TEST_QUANTIZE_LT__(size=[50, 120, 130, 4], iterations=30, sym=True)
__TEST_QUANTIZE_LT__(size=[12, 74, 15, 411], iterations=30, sym=False)
__TEST_QUANTIZE_LT__(size=[50, 7, 130, 1], iterations=30, sym=True)
__TEST_QUANTIZE_LT__(size=[12, 4, 15, 3], iterations=30, sym=False)
__TEST_QUANTIZE_LT__(size=[5011, 7, 7, 1], iterations=30, sym=True)
__TEST_QUANTIZE_LT__(size=[122552, 1, 10, 4], iterations=30, sym=False)
__TEST_QUANTIZE_LT__(size=[10, 10, 124, 47], iterations=30, sym=True)
__TEST_QUANTIZE_LT__(size=[19, 42, 150, 3], iterations=30, sym=False)

__TEST_QUANTIZE_LC__(size=[1, 1, 1, 1], iterations=100, sym=True)
__TEST_QUANTIZE_LC__(size=[1, 1, 1, 1], iterations=100, sym=False)
__TEST_QUANTIZE_LC__(size=[5, 12, 13, 4], iterations=100, sym=True)
__TEST_QUANTIZE_LC__(size=[1, 7, 15, 41], iterations=100, sym=False)
__TEST_QUANTIZE_LC__(size=[50, 120, 130, 4], iterations=30, sym=True)
__TEST_QUANTIZE_LC__(size=[12, 74, 15, 411], iterations=30, sym=False)
__TEST_QUANTIZE_LC__(size=[50, 7, 130, 1], iterations=30, sym=True)
__TEST_QUANTIZE_LC__(size=[12, 4, 15, 3], iterations=30, sym=False)
__TEST_QUANTIZE_LC__(size=[5011, 7, 7, 1], iterations=30, sym=True, c=0)
__TEST_QUANTIZE_LC__(size=[122552, 1, 10, 4], iterations=30, sym=False, c=0)
__TEST_QUANTIZE_LC__(size=[10, 10, 124, 47], iterations=30, sym=True, c=3)
__TEST_QUANTIZE_LC__(size=[19, 42, 150, 3], iterations=30, sym=False, c=3)

__TEST_QUANTIZE_LT_B__(size=[1, 1, 1, 1], iterations=100, sym=True)
__TEST_QUANTIZE_LT_B__(size=[1, 1, 1, 1], iterations=100, sym=False)
__TEST_QUANTIZE_LT_B__(size=[5, 12, 13, 4], iterations=100, sym=True)
__TEST_QUANTIZE_LT_B__(size=[1, 7, 15, 41], iterations=100, sym=False)
__TEST_QUANTIZE_LT_B__(size=[5, 12, 2130, 4], iterations=30, sym=True)
__TEST_QUANTIZE_LT_B__(size=[12, 74, 315, 41], iterations=30, sym=False)
__TEST_QUANTIZE_LT_B__(size=[50, 17, 13, 1], iterations=30, sym=True)
__TEST_QUANTIZE_LT_B__(size=[12, 41, 15, 3], iterations=30, sym=False)
__TEST_QUANTIZE_LT_B__(size=[501, 7, 73, 1], iterations=30, sym=True)
__TEST_QUANTIZE_LT_B__(size=[1222, 12, 10, 4], iterations=30, sym=False)
__TEST_QUANTIZE_LT_B__(size=[10, 104, 14, 47], iterations=30, sym=True)
__TEST_QUANTIZE_LT_B__(size=[19, 42, 120, 3], iterations=30, sym=False)

__TEST_QUANTIZE_LC_B__(size=[1, 1, 1, 1], iterations=100, sym=True)
__TEST_QUANTIZE_LC_B__(size=[1, 1, 1, 1], iterations=100, sym=False)
__TEST_QUANTIZE_LC_B__(size=[5, 12, 14, 12], iterations=100, sym=True)
__TEST_QUANTIZE_LC_B__(size=[1, 7, 15, 41], iterations=100, sym=False)
__TEST_QUANTIZE_LC_B__(size=[5, 12, 130, 4], iterations=30, sym=True)
__TEST_QUANTIZE_LC_B__(size=[12, 74, 15, 41], iterations=30, sym=False)
__TEST_QUANTIZE_LC_B__(size=[50, 7, 13, 1], iterations=30, sym=True)
__TEST_QUANTIZE_LC_B__(size=[12, 4, 15, 3], iterations=30, sym=False)
__TEST_QUANTIZE_LC_B__(size=[501, 7, 7, 1], iterations=30, sym=True, c=0)
__TEST_QUANTIZE_LC_B__(size=[1222, 1, 10, 4], iterations=30, sym=False, c=0)
__TEST_QUANTIZE_LC_B__(size=[10, 10, 14, 47], iterations=30, sym=True, c=3)
__TEST_QUANTIZE_LC_B__(size=[19, 42, 10, 3], iterations=30, sym=False, c=3)

# Test Histogram
for i in range(TEST_TIMES):
    t = torch.rand(size=[128, 3, 224, 224]).to(EXECUTING_DEVICE)
    s = 0.01

    # torch hist
    hist = torch.histc(torch.abs(t), bins=50, min=0, max=0.5)

    # t = t.int()
    cuda_hist = torch.zeros(size=[50]).to(EXECUTING_DEVICE).int()
    cuda_hist = CUDA.Histogram_T(t, cuda_hist, s)
    assert torch.abs(hist - cuda_hist).max().item() < 100



# This Example shows you how to execute a quantized network and get its result.
from typing import Iterable

import torch
import torchvision
from ppq import (BaseGraph, QuantableOperation, QuantizationSettingFactory,
                 TargetPlatform, TorchExecutor)
from ppq.api import quantize_torch_model
from torch.utils.data import DataLoader
from tqdm import tqdm

BATCHSIZE = 32
INPUT_SHAPE = [3, 224, 224]
DEVICE = 'cuda' # only cuda is fully tested :(  For other executing device there might be bugs.
PLATFORM = TargetPlatform.PPL_CUDA_INT8 # identify a target platform for your network.

def load_calibration_dataset() -> Iterable:
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]

def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)

# Load a pretrained mobilenet v2 model
model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# create a setting for quantizing your network with PPL CUDA.
quant_setting = QuantizationSettingFactory.pplcuda_setting()
quant_setting.equalization = True # use layerwise equalization algorithm.
quant_setting.dispatcher   = 'conservative' # dispatch this network in conservertive way.

# Load training data for creating a calibration dataloader.
calibration_dataset = load_calibration_dataset()
calibration_dataloader = DataLoader(
    dataset=calibration_dataset,
    batch_size=BATCHSIZE, shuffle=True)

# quantize your model.
quantized = quantize_torch_model(
    model=model, calib_dataloader=calibration_dataloader,
    calib_steps=32, input_shape=[BATCHSIZE] + INPUT_SHAPE,
    setting=quant_setting, collate_fn=collate_fn, platform=PLATFORM,
    onnx_export_file='Output/onnx.model', device=DEVICE, verbose=0)

# Quantization Result is a PPQ BaseGraph instance.
assert isinstance(quantized, BaseGraph)

# build an executor:
executor = TorchExecutor(graph=quantized, device=DEVICE)

# run with your network, results are torch.Tensors
for data in tqdm(calibration_dataloader, desc='Running with executor.'):
    results = executor.forward(inputs=data.to(DEVICE))

# extract result for specific variables:
interested_vars = []
for operation in quantized.operations.values():
    if isinstance(operation, QuantableOperation) and operation.type == 'Conv':
        interested_vars.append(operation.outputs[0].name)

# results contains all convolution layers' output.
results = executor.forward(inputs=data.to(DEVICE), output_names=interested_vars)
print(f'There are {len(results)} convolution results.')



from .baking import ParameterBakingPass
from .base import (QuantizationOptimizationPass,
                   QuantizationOptimizationPipeline)
from .calibration import (IsotoneCalibrationPass, PPLDSPTIReCalibrationPass,
                          RuntimeCalibrationPass)
from .equalization import (ActivationEqualizationPass, ChannelwiseSplitPass,
                           LayerwiseEqualizationPass)
from .extension import ExtensionPass
from .legacy import AdaroundPass
from .morph import (GRUSplitPass, HorizontalLayerSplitPass, MetaxGemmSplitPass,
                    NCNNFormatGemmPass, NXPResizeModeChangePass)
from .parameters import ParameterQuantizePass, PassiveParameterQuantizePass
from .refine import (MishFusionPass, NxpInputRoundingRefinePass,
                     NxpQuantizeFusionPass, QuantAlignmentPass,
                     QuantizeFusionPass, QuantizeSimplifyPass, SwishFusionPass)
from .ssd import SSDEqualizationPass
from .training import BiasCorrectionPass, LearnedStepSizePass, RoundTuningPass



from abc import ABCMeta, abstractmethod
from typing import Any

from ppq.core import (QuantizationStates, TensorQuantizationConfig,
                      ppq_debug_function)
from ppq.IR import Variable


class BaseTensorObserver(metaclass=ABCMeta):
    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        self._watch_on = watch_on
        self._quant_cfg = quant_cfg

    @ abstractmethod
    def observe(self, value: Any):
        raise NotImplementedError('Implement this function first.')

    @ abstractmethod
    def render_quantization_config(self):
        raise NotImplementedError('Implement this function first.')

    def __str__(self) -> str:
        return 'PPQ Tensor Observer (' + self.__class__.__name__ + ') mount on variable ' + \
            self._watch_on.name + ' observing algorithm: ' + self._quant_cfg.observer_algorithm

    @ ppq_debug_function
    def report(self) -> str:
        if self._quant_cfg.state == QuantizationStates.ACTIVATED:
            return f'Observer on Variable {self._watch_on.name}, '\
                f'computed scale: {self._quant_cfg.scale}, computed offset: {self._quant_cfg.offset}\n'



from ppq.core.quant import RoundingPolicy
from ppq.utils.round import ppq_numerical_round, ppq_round_to_power_of_2

if __name__ == '__main__':
    assert ppq_numerical_round(1.5, policy=RoundingPolicy.ROUND_HALF_EVEN) == 2
    assert ppq_numerical_round(2.5, policy=RoundingPolicy.ROUND_HALF_EVEN) == 2
    assert ppq_numerical_round(0.5, policy=RoundingPolicy.ROUND_HALF_EVEN) == 0
    assert ppq_numerical_round(-0.5, policy=RoundingPolicy.ROUND_HALF_EVEN) == 0
    assert ppq_numerical_round(1.1, policy=RoundingPolicy.ROUND_HALF_EVEN) == 1
    assert ppq_numerical_round(1.2, policy=RoundingPolicy.ROUND_HALF_EVEN) == 1
    assert ppq_numerical_round(1.3, policy=RoundingPolicy.ROUND_HALF_EVEN) == 1
    assert ppq_numerical_round(-1.1, policy=RoundingPolicy.ROUND_HALF_EVEN) == -1
    assert ppq_numerical_round(-1.2, policy=RoundingPolicy.ROUND_HALF_EVEN) == -1
    assert ppq_numerical_round(-1.3, policy=RoundingPolicy.ROUND_HALF_EVEN) == -1

    assert ppq_numerical_round(1.5, policy=RoundingPolicy.ROUND_HALF_UP) == 2
    assert ppq_numerical_round(2.5, policy=RoundingPolicy.ROUND_HALF_UP) == 3
    assert ppq_numerical_round(0.5, policy=RoundingPolicy.ROUND_HALF_UP) == 1
    assert ppq_numerical_round(-0.5, policy=RoundingPolicy.ROUND_HALF_UP) == 0

    assert ppq_numerical_round(1.5, policy=RoundingPolicy.ROUND_HALF_DOWN) == 1
    assert ppq_numerical_round(2.5, policy=RoundingPolicy.ROUND_HALF_DOWN) == 2
    assert ppq_numerical_round(0.5, policy=RoundingPolicy.ROUND_HALF_DOWN) == 0
    assert ppq_numerical_round(-0.5, policy=RoundingPolicy.ROUND_HALF_DOWN) == -1

    assert ppq_numerical_round(1.5, policy=RoundingPolicy.ROUND_HALF_TOWARDS_ZERO) == 1
    assert ppq_numerical_round(2.5, policy=RoundingPolicy.ROUND_HALF_TOWARDS_ZERO) == 2
    assert ppq_numerical_round(0.5, policy=RoundingPolicy.ROUND_HALF_TOWARDS_ZERO) == 0

    # 我并不知道下面这个判断为什么会出错
    # assert ppq_numerical_round(-0.5, policy=RoundingPolicy.ROUND_HALF_TOWARDS_ZERO) == 0

    assert ppq_round_to_power_of_2(1.0) == 1
    assert ppq_round_to_power_of_2(1.2) == 2
    assert ppq_round_to_power_of_2(3.2) == 4
    assert ppq_round_to_power_of_2(0.26) == 0.5
    assert ppq_round_to_power_of_2(0.24) == 0.25



import onnxruntime
from ppq.IR.morph import GraphFormatter
from tests.tmodel  import *
from tests.tscheme import *
from ppq     import *
from ppq.api import *
from ppq import layerwise_error_analyse
import sys

DEVICE = 'cuda'
PLATFORM = TargetPlatform.ORT_OOS_INT8

for case in TORCH_TEST_CASES:
    try:
        print(f'PPQ System test(Onnxruntime) start with model {case.model_name}')
        dataset = [case.input_generator().to(DEVICE) for _ in range(8)]
        model = case.model_builder().to(DEVICE)

        quantized = quantize_torch_model(
            model=model,
            calib_dataloader=dataset,
            calib_steps=8,
            input_shape=case.input_generator().shape,
            platform=PLATFORM,
            setting=QuantizationSettingFactory.default_setting())

        executor = TorchExecutor(quantized)
        sample_output = [(executor(sample)[0].cpu().unsqueeze(0)) for sample in dataset]

        export_ppq_graph(
            graph=quantized,
            platform=TargetPlatform.ONNXRUNTIME,
            graph_save_to='onnxruntime',
            config_save_to='export.json')

        # graph has only 1 input and output.
        for name in quantized.outputs: output_name = name
        for name in quantized.inputs: input_name = name
        session = onnxruntime.InferenceSession('onnxruntime.onnx', providers=onnxruntime.get_available_providers())
        onnxruntime_outputs = [session.run([output_name], {input_name: convert_any_to_numpy(sample)}) for sample in dataset]
        onnxruntime_outputs = [convert_any_to_torch_tensor(sample) for sample in onnxruntime_outputs]

        error = []
        for ref, real in zip(sample_output, onnxruntime_outputs):
            error.append(torch_snr_error(ref, real))
        error = sum(error) / len(error) * 100

        print(f'Simulating Error: {error: .4f}%')
        assert error < 1
    except NotImplementedError as e:
        print(f'{time.strftime("%Y-%m-%d %H:%M:%S")} | Error occurred: {e}')
        sys.exit(1)

for case in TORCH_TEST_CASES:
    try:
        print(f'PPQ System test(Onnxruntime) start with model {case.model_name}')
        dataset = [case.input_generator().to(DEVICE) for _ in range(8)]
        model = case.model_builder().to(DEVICE)

        quantized = quantize_torch_model(
            model=model,
            calib_dataloader=dataset,
            calib_steps=8,
            input_shape=case.input_generator().shape,
            platform=PLATFORM,
            setting=QuantizationSettingFactory.default_setting())

        '''
        quantized.outputs.clear()
        quantized.mark_variable_as_graph_output(quantized.variables['onnx::Conv_26'])
        processor = GraphFormatter(quantized)
        processor.truncate_on_var(quantized.variables['onnx::Conv_26'], mark_as_output=True)
        '''

        executor = TorchExecutor(quantized)
        sample_output = [(executor(sample)[0].cpu().unsqueeze(0)) for sample in dataset]

        export_ppq_graph(
            graph=quantized,
            platform=TargetPlatform.ORT_OOS_INT8,
            graph_save_to='onnx',
            config_save_to='export.json')

        # graph has only 1 input and output.
        for name in quantized.outputs: output_name = name
        for name in quantized.inputs: input_name = name
        session = onnxruntime.InferenceSession('onnx.onnx', providers=onnxruntime.get_available_providers())
        onnxruntime_outputs = [session.run([output_name], {input_name: convert_any_to_numpy(sample)}) for sample in dataset]
        onnxruntime_outputs = [convert_any_to_torch_tensor(sample) for sample in onnxruntime_outputs]

        error = []
        for ref, real in zip(sample_output, onnxruntime_outputs):
            error.append(torch_snr_error(ref, real))
        error = sum(error) / len(error) * 100

        print(f'Simulating Error: {error: .4f}%')
        # assert error < 1

    except NotImplementedError as e:
        print(f'{time.strftime("%Y-%m-%d %H:%M:%S")} | Error occurred: {e}')
        sys.exit(1)



import torch
from ppq.core import QuantizationProperty, TensorQuantizationConfig

from .base import BaseQuantFunction
from .floating import PPQFloatingQuantFunction
from .linear import (PPQDyamicLinearQuantFunction, PPQLinearQuant_toInt,
                     PPQLinearQuantFunction)


def PPQuantFunction(tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    """
    ## PPQ 核心量化函数

    根据 config 中描述的策略，量化给定的 tensor.
    
    请注意 config.state 必须处在激活状态，该函数起作用。如果 config.state 处于
        INITIAL, FP32, PASSIVE_INIT 等未激活状态，该函数不进行任何处理，直接返回 tensor.

    ### 线性量化(QuantizationProperty.LINEAR):

        INT8 = Clip(Round((FP32 / scale)))

    ### 浮点量化(QuantizationProperty.FLOATING):

        FP8 = Clip(FP32_TO_FP8((FP32 / scale)))

    ### 动态线性量化(QuantizationProperty.DYNMAIC)

        scale = max(FP32) / 255

        INT8  = Clip(Round((FP32 / scale)))

    """
    if tensor is None: raise ValueError('Tensor is empty.')
    if config.policy.has_property(QuantizationProperty.LINEAR):
        if not config.policy.has_property(QuantizationProperty.DYNAMIC):
            return PPQLinearQuantFunction(tensor, config)
        else: return PPQDyamicLinearQuantFunction(tensor, config)

    if config.policy.has_property(QuantizationProperty.FLOATING):
        return PPQFloatingQuantFunction(tensor, config)

    raise ValueError('Unexpected Quantization Property Found in PPQuantFunction. '
                     'Do not konw how to quantize your config yet.')


def PPQuantFunction_toInt(tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    """
    ## PPQ 核心量化函数

    根据 config 中描述的策略，这个函数将会执行线性量化，动态量化
    
    但是结果直接出来是整数
    """

    if config.policy.has_property(QuantizationProperty.LINEAR):
        if not config.policy.has_property(QuantizationProperty.DYNAMIC):
            return PPQLinearQuant_toInt(tensor, config)

    raise ValueError('Unexpected Quantization Property Found in PPQuantFunction_toInt. '
                     'Do not konw how to quantize your config yet.')


__all__ = ['PPQuantFunction', 'PPQuantFunction_toInt', 'PPQDyamicLinearQuantFunction',
           'PPQLinearQuantFunction', 'PPQFloatingQuantFunction', 'BaseQuantFunction', 
           'PPQLinearQuant_toInt']



from .default import (DEFAULT_BACKEND_TABLE, AveragePool_forward, AdaptiveAvgPool2d_forward,
                      BatchNormalization_forward, Cast_forward, Clip_forward,
                      Concat_forward, Constant_forward,
                      ConstantOfShape_forward, Conv_forward, Eltwise_forward,
                      Equal_forward, Expand_forward, Flatten_forward,
                      Gather_forward, GatherND_forward, Gemm_forward,
                      Greater_forward, MaxPool2d_forward, _NMS_forward,
                      NonZero_forward, Range_forward, ReduceL2_forward,
                      ReduceMax_forward, Reshape_forward, Resize_forward,
                      ScatterElements_forward, ScatterND_forward,
                      Shape_forward, Slice_forward, Softmax_forward,
                      Squeeze_forward, Tile_forward, TopK_forward,
                      Transpose_forward, UnaryEltwise_forward,
                      Unsqueeze_forward, Where_forward)
from .dsp import PPL_DSP_BACKEND_TABLE
from .cuda import PPL_GPU_BACKEND_TABLE
from .nxp import NXP_BACKEND_TABLE
from .extension import EXTENSION_BACKEND_TABLE
from .base import TorchBackendContext
from .onnx import ONNX_BACKEND_TABLE
from .academic import ACADEMIC_BACKEND_TABLE



import torch
from ppq.core import (PPQ_CONFIG, QuantizationProperty, QuantizationStates,
                      RoundingPolicy, TensorQuantizationConfig)
from torch.autograd import Function


class TensorwiseFloatingQuantImpl(Function):
    """Torch Tensorwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will use ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """
    @ staticmethod
    def forward(ctx, tensor: torch.Tensor, scales: torch.Tensor, offsets: torch.Tensor,
                exponet_bits: int, mantissa_bits: int,
                quant_min: float, quant_max: float,
                rounding: RoundingPolicy) -> torch.Tensor:

        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)
        if not PPQ_CONFIG.USING_CUDA_KERNEL or not tensor.is_cuda:
            # quantization function, pytorch implementation
            raise NotImplementedError('This Feature must run with PPQ Cuda Kernel.')
        
        else:
            from ppq.core import CUDA

            # quantization function, pure cuda implementation
            quantized = CUDA.FloatingQuantize_T(
                tensor=tensor,
                scales=scales,
                offsets=offsets,
                exponent=exponet_bits,
                mantissa=mantissa_bits,
                minimum=quant_min,
                maximum=quant_max,
                rounding=rounding.value
            )
            return quantized

    @ staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None, None, None, None, None, None, None, None


class ChannelwiseFloatingQuantImpl(Function):
    """Torch Channelwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will bypass this function by using ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """
    @ staticmethod
    def forward(ctx, tensor: torch.Tensor, scales: torch.Tensor,
                offsets: torch.Tensor, channel_axis: int,
                exponet_bits: int, mantissa_bits: int, 
                quant_min: float, quant_max: float,
                rounding: RoundingPolicy) -> torch.Tensor:

        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)
        if not PPQ_CONFIG.USING_CUDA_KERNEL or not tensor.is_cuda:
            # generate a shape that likes [1, 1, -1, 1], the only -1 is at channel axe.
            raise NotImplementedError('This Feature must run with PPQ Cuda Kernel.')
        else:
            from ppq.core import CUDA
            quantized = CUDA.FloatingQuantize_C(
                tensor=tensor,
                scales=scales,
                offsets=offsets,
                channel_axis=channel_axis,
                exponent=exponet_bits,
                mantissa=mantissa_bits,
                minimum=quant_min,
                maximum=quant_max,
                rounding=rounding.value)
            return quantized

    @ staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None, None, None, None, None, None, None, None, None


def PPQFloatingQuantFunction(
    tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    if not PPQ_CONFIG.USING_CUDA_KERNEL:
        raise PermissionError('PPQ Floating Quant Function require PPQ_CONFIG.USING_CUDA_KERNEL = True')
    if not tensor.is_cuda:
        raise PermissionError('PPQ Floating Quant Function requires tensor device to be cuda, '
                              'CPU floating quantization is not implemented yet.')

    """PPQ 核心量化函数，没啥好说的了吧，这个玩意既做 quant 也做 dequant"""
    if not QuantizationStates.is_activated(config.state): return tensor
    if not config.policy.has_property(QuantizationProperty.FLOATING):
        raise ValueError('Critical Quantization Error! Unexpected policy detected. '
                         'PPQFloatingQuantFunction except a Floating Quantization Config.')
    if config.policy.has_property(QuantizationProperty.DYNAMIC):
        raise ValueError('Unexpected Dynamic Flag in Quantization Policy.')

    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        return ChannelwiseFloatingQuantImpl.apply(
            tensor, config.scale, config.offset, config.channel_axis,
            config.exponent_bits, config.mantissa_bits,
            config.quant_min, config.quant_max, config.rounding)
    elif config.policy.has_property(QuantizationProperty.PER_TENSOR):
        return TensorwiseFloatingQuantImpl.apply(
            tensor, config.scale, config.offset,
            config.exponent_bits, config.mantissa_bits,
            config.quant_min, config.quant_max, config.rounding)



from typing import Union

import torch
from ppq.api.setting import QuantizationSetting
from ppq.core import (PASSIVE_OPERATIONS,
                      OperationQuantizationConfig, QuantizationPolicy,
                      QuantizationProperty, QuantizationStates, RoundingPolicy,
                      TargetPlatform)
from ppq.executor.base import BaseGraphExecutor
from ppq.IR import BaseGraph, Operation
from ppq.quantization.optim.base import QuantizationOptimizationPipeline
from ppq.quantization.optim.morph import MetaxGemmSplitPass

from .base import BaseQuantizer


class MetaxTensorwiseQuantizer(BaseQuantizer):
    def __init__(
        self,
        graph: BaseGraph
    ) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = 0
        self._quant_max = 255

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:

        base_quant_config = self.create_default_quant_config(
            op=operation, num_of_bits=self._num_of_bits, exponent_bits=0,
            quant_max=self._quant_max, quant_min=self._quant_min,
            observer_algorithm='percentile', policy=self.quantize_policy,
            rounding=self.rounding_policy,
        )

        if operation.type in {'Conv'}:
            # if operation has bias
            if operation.num_of_input == 3:
                bias_config = base_quant_config.input_quantization_config[-1]
                # bias should be quantized with 32 bits
                # in python3, int indicates long long in C++
                # so that it has enough precision to represent a number like 2^32
                # however, it may cause a scale underflow
                # here we give bias a 30 bits precision, which is pettery enough in all cases
                bias_config.num_of_bits = 30
                bias_config.quant_max = int(pow(2, 30 - 1))
                bias_config.quant_min = - int(pow(2, 30 - 1))
                bias_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.LINEAR +
                    QuantizationProperty.PER_TENSOR)
                bias_config.state = QuantizationStates.FP32
            for config in base_quant_config.input_quantization_config[1: ]:
                config.observer_algorithm = 'minmax'

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False
        return base_quant_config

    @ property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.METAX_INT8_T

    @ property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @ property
    def quant_operation_types(self) -> set:
        return {
            'Conv', 'Relu', 'PRelu', 'Clip', 'Gemm',
            'Resize', 'MaxPool', 'AveragePool',
            'GlobalMaxPool', 'GlobalAveragePool',
            'Mul', 'Add', 'LeakyRelu', 'Split', 'Concat',
            'Transpose', 'Slice', 'Reshape', 'Flatten',
            'MatMul'}

    @ property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.ASYMMETRICAL +
            QuantizationProperty.LINEAR +
            QuantizationProperty.PER_TENSOR)

    @ property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    def build_prequant_pipeline(
        self, setting: QuantizationSetting, executor: BaseGraphExecutor) -> QuantizationOptimizationPipeline:
        return super().build_prequant_pipeline(setting, executor).append_optimization_to_pipeline(
            MetaxGemmSplitPass(), at_front=True)

    @ property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip'}


class MetaxChannelwiseQuantizer(BaseQuantizer):
    def __init__(
        self, graph: Union[BaseGraph, BaseGraph]
    ) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = -127
        self._quant_max = 127

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy, rounding=self.rounding_policy,
            op=operation, num_of_bits=self._num_of_bits, exponent_bits=0,
            quant_max=self._quant_max, quant_min=self._quant_min,
            observer_algorithm='percentile'
        )

        if operation.type in {'Conv', 'MatMul'}:
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel quant-config.
            assert operation.num_of_input > 0, 'Seems you got a Conv layer with no parameters.'

            # first parameter must exits, for conv layer it will be conv_weight
            # layout: [out_channel, in_channel, kernel_size, kernel_size]
            if operation.type in {'Conv'}:
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.quant_max = 127
                conv_weight_config.quant_min = -127
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.LINEAR +
                    QuantizationProperty.PER_CHANNEL
                )
                conv_weight_config.channel_axis = 0
                conv_weight_config.observer_algorithm = 'Minmax'
            # first parameter must exits, for gemm layer it will be gemm_weight
            # layout: [in_dim, out_dim]
            elif operation.type in {'MatMul'}:
                matmul_weight_config = base_quant_config.input_quantization_config[1]
                matmul_weight_config.quant_max = 127
                matmul_weight_config.quant_min = -127
                matmul_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.LINEAR +
                    QuantizationProperty.PER_CHANNEL
                )
                matmul_weight_config.channel_axis = 1
                matmul_weight_config.observer_algorithm = 'Minmax'
            # if operation has bias
            if operation.type == 'Conv' and operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.LINEAR +
                    QuantizationProperty.PER_CHANNEL
                )
                bias_config.num_of_bits = 30
                bias_config.quant_max = int(pow(2, bias_config.num_of_bits - 1))
                bias_config.quant_min = - int(pow(2, bias_config.num_of_bits - 1))
                bias_config.state = QuantizationStates.PASSIVE_INIT
                bias_config.channel_axis = 0
                bias_config.observer_algorithm = 'minmax'

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False
        return base_quant_config

    @ property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.METAX_INT8_C

    @ property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @ property
    def quant_operation_types(self) -> set:
        return {
            'Conv', 'Relu', 'PRelu', 'Clip', 'Gemm',
            'Resize', 'MaxPool', 'AveragePool',
            'GlobalMaxPool', 'GlobalAveragePool',
            'Mul', 'Add', 'LeakyRelu', 'Split', 'Concat',
            'Transpose', 'Slice', 'Reshape', 'Flatten',
            'MatMul'}

    @ property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL +
            QuantizationProperty.LINEAR +
            QuantizationProperty.PER_TENSOR
        )

    @ property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    def build_prequant_pipeline(
        self, setting: QuantizationSetting, executor: BaseGraphExecutor) -> QuantizationOptimizationPipeline:
        return super().build_prequant_pipeline(setting, executor).append_optimization_to_pipeline(
            MetaxGemmSplitPass(), at_front=True)

    @ property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip'}



# This Example shows you how to replace quantization function with your own quant delegator.

from typing import Iterable

import torch
import torchvision
from ppq import *
from ppq.api import quantize_torch_model
from torch.utils.data import DataLoader
from tqdm import tqdm

class MyQuantDelegator(TorchQuantizeDelegator):
    """Use This class to realize your quantization logic.

    Inherit class TorchQuantizeDelegate, implement interface __call__, then
    register your delegator with executor.register_quantize_delegate
    """
    def __call__(self, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
            raise ValueError('Sorry, this delegator handles only Symmetrical Quantizations.')
        print('You are invoking cusitmized quant function now.')
        return torch.round(tensor / config.scale) * config.scale

BATCHSIZE = 32
INPUT_SHAPE = [3, 224, 224]
DEVICE = 'cuda' # only cuda is fully tested :(  For other executing device there might be bugs.
PLATFORM = TargetPlatform.PPL_CUDA_INT8 # identify a target platform for your network.

def load_calibration_dataset() -> Iterable:
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]

def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)

# Load a pretrained mobilenet v2 model
model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# create a setting for quantizing your network with PPL CUDA.
quant_setting = QuantizationSettingFactory.pplcuda_setting()
quant_setting.equalization = True # use layerwise equalization algorithm.
quant_setting.dispatcher   = 'conservative' # dispatch this network in conservertive way.

# Load training data for creating a calibration dataloader.
calibration_dataset = load_calibration_dataset()
calibration_dataloader = DataLoader(
    dataset=calibration_dataset,
    batch_size=BATCHSIZE, shuffle=True)

# quantize your model.
quantized = quantize_torch_model(
    model=model, calib_dataloader=calibration_dataloader,
    calib_steps=32, input_shape=[BATCHSIZE] + INPUT_SHAPE,
    setting=quant_setting, collate_fn=collate_fn, platform=PLATFORM,
    onnx_export_file='Output/onnx.model', device=DEVICE, verbose=0)

# Quantization Result is a PPQ BaseGraph instance.
assert isinstance(quantized, BaseGraph)

# build an executor:
executor = TorchExecutor(graph=quantized, device=DEVICE)

# register quant delegator
for var in quantized.variables.values():
    if isinstance(var, QuantableVariable):
        if var.source_op_config is not None:
            executor.register_quantize_delegate(var.source_op_config, MyQuantDelegator)

# run with your network, results are torch.Tensors
for data in tqdm(calibration_dataloader, desc='Running with executor.'):
    results = executor.forward(inputs=data.to(DEVICE))
    break

# remove delegators
for var in quantized.variables.values():
    if isinstance(var, QuantableVariable):
        if var.source_op_config is not None:
            executor.remove_quantize_delegate(var.source_op_config)



# Test Quantization System Performace on Image Classification Models with ILSVRC2012 Dataset
#
#   1. How to use: 
#      Run this script with python directly.

# Quantizer Configuration
SYMMETRICAL = True
POWER_OF_2  = True
PER_CHANNEL = False
FP8         = True
BIT_WIDTH   = 8

# Should contains model file(.onnx)
MODEL_DIR = 'QuantZoo/Model/Imagenet'

# Should contains Calib & Test Img Folder
CALIB_DIR = 'QuantZoo/Data/Imagenet/Calib'
TEST_DIR  = 'QuantZoo/Data/Imagenet/Test'

# calibration & test batchsize
BATCHSIZE = 32

# write report to here
REPORT_DIR = 'QuantZoo/Reports'

CONFIGS = [
{
    'Model': 'repvgg',
    'Output': ['input.172']
},
{
    'Model': 'efficientnet_v1_b0',
    'Output': ['/features/features.8/features.8.2/Mul_output_0']
},
{
    'Model': 'efficientnet_v1_b1',
    'Output': ['/features/features.8/features.8.2/Mul_output_0']
},
{
    'Model': 'efficientnet_v2_s',
    'Output': ['/features/features.7/features.7.2/Mul_output_0']
},
{
    'Model': 'mnasnet0_5',
    'Output': ['/layers/layers.16/Relu_output_0']
},
{
    'Model': 'mnasnet1_0',
    'Output': ['/layers/layers.16/Relu_output_0']
},
{
    'Model': 'mobilenet_v2',
    'Output': ['/features/features.18/features.18.2/Clip_output_0']
},
{
    'Model': 'resnet18',
    'Output': ['/layer4/layer4.1/relu_1/Relu_output_0']
},
{
    'Model': 'resnet50',
    'Output': ['/layer4/layer4.2/relu_2/Relu_output_0']
},

{
    'Model': 'mobilenet_v3_large',
    'Output': ['/classifier/classifier.1/Mul_output_0']
},
{
    'Model': 'mobilenet_v3_small',
    'Output': ['/classifier/classifier.1/Mul_output_0']
},
{
    'Model': 'v100_gpu64@5ms_top1@71.6_finetune@25',
    'Output': ['471']
},
{
    'Model': 'v100_gpu64@6ms_top1@73.0_finetune@25',
    'Output': ['471']
},
{
    'Model': 'lcnet_050',
    'Output': ['/act2/Mul_output_0']
},
{
    'Model': 'lcnet_100',
    'Output': ['/act2/Mul_output_0']
},


{
    # vit_b_16 requires BATCHSIZE = 1!
    'Model': 'vit_b_16',
    'Output': ['onnx::Gather_1703']
},
]

import os

import torch

import ppq.lib as PFL
from ppq.api import ENABLE_CUDA_KERNEL, load_onnx_graph
from ppq.core import TargetPlatform
from ppq.executor import TorchExecutor
from ppq.quantization.optim import (LayerwiseEqualizationPass,
                                    LearnedStepSizePass, ParameterQuantizePass,
                                    RuntimeCalibrationPass)
from QuantZoo.Data.Imagenet.Eval import (evaluate_ppq_module_with_imagenet,
                                         load_imagenet_from_directory)
from QuantZoo.Quantizers import MyFP8Quantizer, MyInt8Quantizer
from QuantZoo.Util import error_analyze


calib_loader = load_imagenet_from_directory(
    directory=CALIB_DIR, batchsize=BATCHSIZE,
    shuffle=False, require_label=False,
    num_of_workers=8)


test_loader = load_imagenet_from_directory(
    directory=TEST_DIR, batchsize=BATCHSIZE,
    shuffle=False, require_label=True, 
    num_of_workers=8)


with ENABLE_CUDA_KERNEL():
    for config in CONFIGS:
        model = config['Model']
        monitoring_vars = config['Output']

        print(f"Ready to run quant benchmark on {model}")
        graph = load_onnx_graph(onnx_import_file=os.path.join(MODEL_DIR, model + '.onnx'))

        if model == 'vit_b_16':
            if BATCHSIZE == 32:
                raise Exception('To Evaluate vit_b_16, change batchsize to 1, change calibration method to minmax.')
            from ppq.IR import GraphMerger
            processor = GraphMerger(graph)
            processor.fuse_matmul_add()
            processor.fuse_layernorm()
            processor.fuse_gelu()

        quantizer = MyInt8Quantizer(
            graph=graph, sym=SYMMETRICAL, power_of_2=POWER_OF_2, 
            num_of_bits=BIT_WIDTH, per_channel=PER_CHANNEL)
        if FP8: quantizer = MyFP8Quantizer(graph=graph, calibration='floating')

        # convert op to quantable-op
        for name, op in graph.operations.items():
            if op.type in {'Conv', 'ConvTranspose', 'MatMul', 'Gemm', 
                           'PPQBiasFusedMatMul', 'LayerNormalization'}:
                quantizer.quantize_operation(name, platform=TargetPlatform.INT8)

        # build quant pipeline.
        pipeline = PFL.Pipeline([
            # LayerwiseEqualizationPass(iteration=10),
            ParameterQuantizePass(),
            RuntimeCalibrationPass(),
            # LearnedStepSizePass(steps=500, collecting_device='cuda', block_size=5)
        ])

        # call pipeline.
        executor = TorchExecutor(graph=graph)
        executor.tracing_operation_meta(torch.zeros(size=[BATCHSIZE, 3, 224, 224]).cuda())

        pipeline.optimize(
            graph=graph, dataloader=calib_loader, verbose=True,
            calib_steps=32, collate_fn=lambda x: x.to('cuda'), executor=executor)

        # evaluation
        acc = evaluate_ppq_module_with_imagenet(
            model=graph, imagenet_validation_loader=test_loader,
            batchsize=BATCHSIZE, device='cuda', verbose=False)
        print(f'Model Classify Accurarcy = {acc: .4f}%')

        # error analyze
        performance = error_analyze(
            graph=graph,
            outputs=monitoring_vars,
            dataloader=test_loader, 
            collate_fn=lambda x: x[0].to('cuda'),
            verbose=True
        )


import numpy as np
from ppq.core.data import (DataType, convert_any_to_numpy,
                           convert_any_to_python_primary_type)
from ppq.IR import Operation
from ppq.log import NaiveLogger

from . import ppl_caffe_pb2

logger = NaiveLogger.get_logger('PPQ')
caffe_export_map = {}


def register_class(cls):
    caffe_export_map[cls.__name__] = cls
    return cls

def refine_value(attribute):
    if isinstance(attribute, DataType):
        attribute = attribute.value
    return convert_any_to_python_primary_type(attribute)

class CaffeOpExporter(object):
    onnx_to_caffe = {
        'Conv': 'Convolution',
        'BatchNormalization': 'BN',
        'Relu': 'ReLU',
        'PRelu': 'PReLU',
        'LeakyRelu': 'ReLU',
        'GlobalAveragePool': 'Pooling',
        'GlobalMaxPool': 'Pooling',
        'MaxPool': 'Pooling',
        'AveragePool': 'Pooling',
        'Mul': 'Eltwise',
        'Add': 'Eltwise',
        'Max': 'Eltwise',
        'Sub': 'Eltwise',
        'Gemm': 'InnerProduct',
        'Pad': 'ReflectionPad',
        'ConvTranspose': 'Deconvolution',
        'InstanceNormalization': 'InstanceNorm',
        'Pow': 'Power',
        'Tanh': 'TanH',
        'CaffeArgMax': 'ArgMax',
        'DepthToSpace': 'SubpixelUp',
        'SpaceToDepth': 'SubpixelDown',
        'HardSwish': 'HSwish',
        'HardSigmoid': 'HSigmoid',
        'MatMul': 'MatMul'
    }

    def __init__(self, op: Operation):
        self.op = op
        self.op_type = self.set_type()
        self.layer = ppl_caffe_pb2.LayerParameter(type=self.op_type, name=self.op.name)

    def set_type(self):
        op_type = self.op.type
        if op_type in CaffeOpExporter.onnx_to_caffe:
            op_type = CaffeOpExporter.onnx_to_caffe[op_type]
        return op_type

    def set_attr(self):
        pass

    def parse(self) -> ppl_caffe_pb2.LayerParameter:
        self.set_attr()
        self.layer.bottom[:] = [var.name for var in self.op.inputs if not var.is_parameter]
        self.layer.top[:] = [var.name for var in self.op.outputs]

        for var in self.op.parameters:
            blob = ppl_caffe_pb2.BlobProto()
            value = var.value
            value = convert_any_to_numpy(value)
    
            shape = var.shape
            dtype = DataType.to_numpy(var.dtype)

            blob.shape.dim.extend(shape)
            blob.data.extend(value.astype(dtype).flat)
            self.layer.blobs.extend([blob])

        return self.layer


@register_class
class Conv(CaffeOpExporter):
    def set_attr(self):
        kernel_h, kernel_w = refine_value(self.op.attributes.get('kernel_shape'))
        stride_h, stride_w = refine_value(self.op.attributes.get('strides', [1, 1]))
        dilations_h, dilations_w = refine_value(self.op.attributes.get('dilations', [1, 1]))
        pads = refine_value(self.op.attributes.get('pads', [0, 0]))
        if len(pads) == 2:
            pad_h, pad_w = pads
        elif len(pads) == 4:
            begin_pad = pads[:2]
            end_pad = pads[2:]
            if begin_pad == end_pad:
                pad_h, pad_w = begin_pad
            else:
                logger.error('Caffe only support begin_pad == end_pad in layer')
        else:
            logger.error(f'Unsupported pads attributes with the length of {len(pads)} in Caffe')

        self.layer.convolution_param.num_output = self.op.parameters[0].value.shape[0]
        self.layer.convolution_param.group = refine_value(self.op.attributes.get('group', 1))
        self.layer.convolution_param.kernel_h = kernel_h
        self.layer.convolution_param.kernel_w = kernel_w
        self.layer.convolution_param.pad_h = pad_h
        self.layer.convolution_param.pad_w = pad_w
        self.layer.convolution_param.stride_h = stride_h
        self.layer.convolution_param.stride_w = stride_w
        self.layer.convolution_param.hole_h = dilations_h
        self.layer.convolution_param.hole_w = dilations_w

        if len(self.op.parameters) == 2:
            self.layer.convolution_param.bias_term = True
        else:
            self.layer.convolution_param.bias_term = False


@register_class
class BatchNormalization(CaffeOpExporter):
    def set_attr(self):
        self.layer.bn_param.moving_average = bool(refine_value(self.op.attributes.get('training_mode', 0)))
        self.layer.bn_param.var_eps = refine_value(self.op.attributes.get('epsilon', 1e-05))
        if self.layer.bn_param.moving_average:
            self.layer.bn_param.decay = 1 - refine_value(self.op.attributes.get('momentum', 0.9))

    def parse(self):
        super(BatchNormalization, self).parse()
        channel = len(self.layer.blobs[3].data)
        for i in range(4):
            self.layer.blobs[i].shape.ClearField('dim')
            self.layer.blobs[i].shape.dim.extend([1, channel, 1, 1])
        return self.layer


@register_class
class Relu(CaffeOpExporter):
    pass


@register_class
class PRelu(CaffeOpExporter):
    pass


@register_class
class LeakyRelu(CaffeOpExporter):
    def set_attr(self):
        self.layer.relu_param.negative_slope = refine_value(self.op.attributes.get('alpha', 0.01))


class _Pooling(CaffeOpExporter):
    def set_attr(self):
        kernel_h, kernel_w = refine_value(self.op.attributes.get('kernel_shape'))
        stride_h, stride_w = refine_value(self.op.attributes.get('strides', [1, 1]))
        ceil_mode = refine_value(self.op.attributes.get('ceil_mode', 0))
        pads = refine_value(self.op.attributes.get('pads', [0, 0]))
        if len(pads) == 2:
            pad_h, pad_w = pads
        elif len(pads) == 4:
            begin_pad = pads[:2]
            end_pad = pads[2:]
            if begin_pad == end_pad:
                pad_h, pad_w = begin_pad
            else:
                logger.error('Caffe only support begin_pad == end_pad in layer')
        else:
            logger.error(f'Unsupported pads attributes with the length of {len(pads)} in Caffe')

        self.layer.pooling_param.kernel_h = kernel_h
        self.layer.pooling_param.kernel_w = kernel_w
        self.layer.pooling_param.pad_h = pad_h
        self.layer.pooling_param.pad_w = pad_w
        self.layer.pooling_param.stride_h = stride_h
        self.layer.pooling_param.stride_w = stride_w
        if ceil_mode == 0:
            # ceil_mode is True by CaffeOpExporter in caffe
            self.layer.pooling_param.ceil_mode = False


@register_class
class GlobalAveragePool(_Pooling):
    def set_attr(self):
        self.layer.pooling_param.global_pooling = True
        self.layer.pooling_param.pool = ppl_caffe_pb2.PoolingParameter.AVE


@register_class
class AveragePool(_Pooling):
    def set_attr(self):
        super(AveragePool, self).set_attr()
        self.layer.pooling_param.global_pooling = False
        self.layer.pooling_param.pool = ppl_caffe_pb2.PoolingParameter.AVE


@register_class
class GlobalMaxPool(_Pooling):
    def set_attr(self):
        self.layer.pooling_param.global_pooling = True
        self.layer.pooling_param.pool = ppl_caffe_pb2.PoolingParameter.MAX


@register_class
class MaxPool(_Pooling):
    def set_attr(self):
        super(MaxPool, self).set_attr()
        self.layer.pooling_param.global_pooling = False
        self.layer.pooling_param.pool = ppl_caffe_pb2.PoolingParameter.MAX


@register_class
class Concat(CaffeOpExporter):
    def set_attr(self):
        self.layer.concat_param.axis = refine_value(self.op.attributes['axis'])


@register_class
class Softmax(CaffeOpExporter):
    def set_attr(self):
        axis = refine_value(self.op.attributes.get('axis', -1))
        if not (axis == -1 or axis == len(self.op.inputs[0].shape) - 1):
            logger.warning(f'Converting to caffe Softmax, the axis={axis}, which is not the last axis. '
                           'This may result to incorrect caffe model')
        self.layer.softmax_param.axis = axis


@register_class
class Transpose(CaffeOpExporter):
    def set_attr(self):
        perm = refine_value(self.op.attributes['perm'])
        self.layer.transpose_param.dim.extend(perm)


@register_class
class ReduceL2(CaffeOpExporter):
    def set_attr(self):
        self.layer.reducel2_param.axes = refine_value(self.op.attributes.get('axes'))
        self.layer.reducel2_param.keepdims = refine_value(self.op.attributes.get('keepdims', 1))


@register_class
class ReduceMean(CaffeOpExporter):
    def set_attr(self):
        axis = None
        if 'axis' in self.op.attributes:
            axis = self.op.attributes.get('axis')
        elif 'axes' in self.op.attributes:
            axis = self.op.attributes.get('axes')
        if isinstance(axis, list):
            assert len(axis) == 1, (
                'You are trying to dump a RuduceMean op to caffe, '
                f'however caffe support 1 axis only, your mean operation has {len(axis)} working axis')
            axis = axis[0]
        self.layer.reduce_param.axis = axis

@register_class
class Div(CaffeOpExporter):
    pass


@register_class
class Mul(CaffeOpExporter):
    # TODO: Can optimize some case to Scale + Reshape
    #  (lhs_shape[1] == rhs_shape[1] and all(i == 1 for i in rhs_shape[2:]))
    def set_attr(self):
        self.layer.eltwise_param.operation = ppl_caffe_pb2.EltwiseParameter.PROD

    def parse(self):
        self.set_attr()
        self.layer.bottom[:] = [var.name for var in self.op.inputs if not var.is_parameter]
        self.layer.top[:] = [var.name for var in self.op.outputs]

        for var in self.op.parameters:
            value = convert_any_to_numpy(var.value)
            if value.size != 1:
                raise AttributeError(f'Now don\'t support Mul op with initializer in shape {value.shape} convert to caffe')
            # Mul only has two inputs, thus in this loop means the bottom has only one item
            self.layer.eltwise_param.coeff.append(value.item())

        return self.layer


@register_class
class Add(CaffeOpExporter):
    def set_attr(self):
        self.layer.eltwise_param.operation = ppl_caffe_pb2.EltwiseParameter.SUM
        # ONNX op only support no coeff add now
        # https://github.com/onnx/onnx/blob/master/docs/Operators.md#add
        self.layer.eltwise_param.coeff[:] = [1.0] * len(self.op.inputs)

    def parse(self):
        parameter_layers, extend_bottom = [], []
        for i, var in enumerate(self.op.parameters):
            param_layer = ppl_caffe_pb2.LayerParameter(type='Parameter', name=self.op.name + '_param_' + str(i))
            param_layer.top[:] = [var.name]
            extend_bottom.append(var.name)

            blob = ppl_caffe_pb2.BlobProto()
            value = convert_any_to_numpy(var.value)

            shape = var.shape
            dtype = DataType.to_numpy(var.dtype)

            blob.shape.dim.extend(shape)
            blob.data.extend(value.astype(dtype).flat)
            param_layer.blobs.extend([blob])

            shape_param = param_layer.parameter_param
            if len(shape) == 3:
                shape_param.batch, shape_param.m, shape_param.n = shape
            elif len(shape) == 4:
                shape_param.batch, shape_param.channel, shape_param.height, shape_param.width = shape
            else:
                raise AttributeError(f'Cannot convert {self.op.name} to Eltwise op.')

            parameter_layers.append(param_layer)

        super(Add, self).parse()
        if len(extend_bottom) != 0:
            self.layer.bottom.extend(extend_bottom)
            self.layer.eltwise_param.coeff.extend([1.0] * len(extend_bottom))
        return [*parameter_layers, self.layer]


@register_class
class Max(CaffeOpExporter):
    def set_attr(self):
        self.layer.eltwise_param.operation = ppl_caffe_pb2.EltwiseParameter.MAX


@register_class
class Sub(CaffeOpExporter):
    def set_attr(self):
        self.layer.eltwise_param.operation = ppl_caffe_pb2.EltwiseParameter.SUM
        # ONNX op only support no coeff sub now
        # https://github.com/onnx/onnx/blob/master/docs/Operators.md#Sub
        self.layer.eltwise_param.coeff[:] = [1.0, -1.0]


@register_class
class Reshape(CaffeOpExporter):
    def set_attr(self):
        if len(self.op.inputs) == 0:
            raise AttributeError(f'{self.op.name} has no inputs. Cannot convert to caffe op. '
                        'Please optimize the onnx model.')
        shape = convert_any_to_numpy(self.op.parameters[0].value)
        self.layer.reshape_param.shape.dim.extend(shape)

    def parse(self):
        self.set_attr()
        self.layer.bottom[:] = [var.name for var in self.op.inputs if not var.is_parameter]
        self.layer.top[:] = [var.name for var in self.op.outputs]
        return self.layer


@register_class
class Clip(CaffeOpExporter):
    def parse(self):
        self.layer.bottom[:] = [var.name for var in self.op.inputs if not var.is_parameter]
        self.layer.top[:] = [var.name for var in self.op.outputs]

        min_val = refine_value(self.op.attributes.get('min'))
        max_val = refine_value(self.op.attributes.get('max'))
        if len(self.op.parameters) == 2:
            min_val = convert_any_to_numpy(self.op.parameters[0].value).item()
            max_val = convert_any_to_numpy(self.op.parameters[1].value).item()

        if min_val == 0.0 and max_val == 6.0:
            self.layer.type = 'ReLU6'
        else:
            self.layer.clip_param.min = min_val
            self.layer.clip_param.max = max_val
        return self.layer


@register_class
class Gemm(CaffeOpExporter):
    def parse(self):
        super(Gemm, self).parse()
        # Whether need to add transpose layer
        transpose_layer = None
        if refine_value(self.op.attributes.get('transA', 0)) != 0:
            A = self.op.inputs[0]
            shape = A.shape
            if len(shape) == 2:
                transpose_layer = ppl_caffe_pb2.LayerParameter(type='Transpose', name=self.op.name + '_transposed')
                transpose_layer.bottom[:] = [A.name]
                transpose_layer.top[:] = [A.name + '_trans']
                transpose_layer.transpose_param.dim[:] = [1, 0]
                # Modify InnerProduct input
                self.layer.bottom[:] = [A.name + '_trans']
            else:
                raise ValueError('Cannot support transposed gemm with non-2D input.')

        if self.op.attributes.get('transB', 0) == 0:
            B = convert_any_to_numpy(self.op.parameters[0].value)
            BT_value = np.transpose(B, [1, 0])
            self.layer.blobs[0].shape.dim[:] = BT_value.shape
            self.layer.blobs[0].data[:] = BT_value.astype('float32').flat
        
        self.layer.inner_product_param.axis = self.op.attributes.get('axis', 1)
        self.layer.inner_product_param.num_output = self.layer.blobs[0].shape.dim[0]
        self.layer.inner_product_param.bias_term = True if len(self.op.parameters) == 2 else False

        if transpose_layer is None:
            return self.layer
        else:
            return [transpose_layer, self.layer]


@register_class
class Pad(CaffeOpExporter):
    def set_attr(self):
        mode = refine_value(self.op.attributes.get('mode', 'constant'))
        if mode != 'reflect':
            raise TypeError(f'Unsupport pad mode {mode} in caffe op')

        pads = convert_any_to_numpy(self.op.inputs[1].value) if len(self.op.inputs) > 1 else refine_value(self.op.attributes['pads'])
        if len(pads) == 2:
            pads = [pads[0], 0, pads[1], 0]
        phs, pws, phe, pwe = pads
        assert phs == phe and pws == pwe
        self.layer.pad_param.pad_h = phs
        self.layer.pad_param.pad_w = pws


# TODO: InterP and other cases
@register_class
class Resize(CaffeOpExporter):
    def __init__(self, op):
        self.mode = refine_value(op.attributes.get('mode'))
        self.scales = convert_any_to_numpy(op.inputs[2].value) if len(op.inputs) > 2 else []
        self.sizes = convert_any_to_numpy(op.inputs[-1].value) if len(op.inputs) == 4 else []
        super().__init__(op)

    def set_type(self):
        if self.mode == 'nearest' and len(self.sizes) == 0 and len(self.scales) > 0:
            return 'NNUpsample'
        elif self.mode == 'linear' and len(self.scales) == 0 and len(self.sizes) > 0:
            return 'Interp'
        else:
            raise TypeError(f'Cannot convert {self.op.name} to caffe op')

    def set_attr(self):
        if self.op_type == 'NNUpsample':
            assert len(self.scales) == 4
            valid_flag = (self.scales[0] == 1.0) and (self.scales[1] == 1.0) and (self.scales[2] == self.scales[3])
            if not valid_flag:
                raise AttributeError(f'Cannot convert {self.op.name} to NNUpsample due to different scales')
            self.layer.nn_upsample_param.resize = int(self.scales[2])
        if self.op_type == 'Interp':
            self.layer.interp_param.height = int(self.sizes[-2])
            self.layer.interp_param.width = int(self.sizes[-1])
            trans_mode = refine_value(self.op.attributes.get('coordinate_transformation_mode', 'half_pixel'))
            if trans_mode == 'align_corners':
                self.layer.interp_param.align_corners = True
            elif trans_mode == 'half_pixel':
                self.layer.interp_param.align_corners = False
            else:
                raise AttributeError(f'Cannot convert {self.op.name} in {trans_mode} mode to Interp.')

    def parse(self):
        self.set_attr()
        self.layer.bottom[:] = [var.name for var in self.op.inputs if not var.is_parameter]
        self.layer.top[:] = [var.name for var in self.op.outputs]
        return self.layer


@register_class
class ConvTranspose(CaffeOpExporter):
    def set_attr(self):
        kernel_h, kernel_w = refine_value(self.op.attributes.get('kernel_shape'))
        stride_h, stride_w = refine_value(self.op.attributes.get('strides', [1, 1]))
        dilations_h, dilations_w = refine_value(self.op.attributes.get('dilations', [1, 1]))
        pads = refine_value(self.op.attributes.get('pads', [0, 0]))
        if len(pads) == 2:
            pad_h, pad_w = pads
        elif len(pads) == 4:
            begin_pad = pads[:2]
            end_pad = pads[2:]
            if begin_pad == end_pad:
                pad_h, pad_w = begin_pad
            else:
                logger.error('Caffe only support begin_pad == end_pad in layer')
        else:
            logger.error(f'Unsupported pads attributes with the length of {len(pads)} in Caffe')

        self.layer.convolution_param.num_output = self.op.parameters[0].value.shape[1] * self.op.attributes.get('group', 1)
        self.layer.convolution_param.group = refine_value(self.op.attributes.get('group', 1))
        self.layer.convolution_param.kernel_h = kernel_h
        self.layer.convolution_param.kernel_w = kernel_w
        self.layer.convolution_param.pad_h = pad_h
        self.layer.convolution_param.pad_w = pad_w
        self.layer.convolution_param.stride_h = stride_h
        self.layer.convolution_param.stride_w = stride_w
        self.layer.convolution_param.hole_h = dilations_h
        self.layer.convolution_param.hole_w = dilations_w

        if len(self.op.parameters) == 2:
            self.layer.convolution_param.bias_term = True
        else:
            self.layer.convolution_param.bias_term = False


@register_class
class Sigmoid(CaffeOpExporter):
    pass


@register_class
class Slice(CaffeOpExporter):
    def parse(self):
        # assert (len(self.op.inputs) == 1)
        input_shape = self.op.inputs[0].shape
        starts, ends = convert_any_to_numpy(self.op.parameters[0].value), convert_any_to_numpy(self.op.parameters[1].value)
        axes = convert_any_to_numpy(self.op.parameters[2].value) if len(self.op.parameters) >= 3 else [i for i in range(len(input_shape))]
        if len(self.op.parameters) >= 4 and any(convert_any_to_numpy(self.op.parameters[3].value) != 1):
            raise AttributeError('Slice op with steps cannot dump to caffe model')

        layers = []
        self.layer = ppl_caffe_pb2.LayerParameter(type=self.op_type, name=self.op.name)

        for i, (start_point, end_point, axis) in enumerate(list(zip(starts, ends, axes))):
            current_layer = ppl_caffe_pb2.LayerParameter(type=self.op_type, name=self.op.name + '_' + str(i))
            current_layer.slice_param.axis = axis
            current_layer.bottom[:] = [var.name for var in self.op.inputs if not var.is_parameter]
            current_layer.top[:] = [var.name for var in self.op.outputs]

            slice_points = [start_point, end_point]
            if start_point == 0:
                slice_points.remove(start_point)
            else:
                current_layer.top.insert(0, self.op.outputs[0].name + '_front')
            if end_point == -1 or end_point == input_shape[axis]:
                slice_points.remove(end_point)
            else:
                current_layer.top.append(self.op.outputs[-1].name + '_behind')
            current_layer.slice_param.slice_point.extend(slice_points)
            layers.append(current_layer)

        return layers


@register_class
class Tanh(CaffeOpExporter):
    pass


@register_class
class Pow(CaffeOpExporter):
    def set_attr(self):
        self.layer.power_param.power = refine_value(self.op.attributes.get('power', 1))
        self.layer.power_param.scale = refine_value(self.op.attributes.get('scale', 1))
        self.layer.power_param.shift = refine_value(self.op.attributes.get('shift', 0))


@register_class
class Scale(CaffeOpExporter):
    def set_attr(self):
        self.layer.scale_param.axis = refine_value(self.op.attributes.get('axis', 1))
        self.layer.scale_param.num_axes = refine_value(self.op.attributes.get('num_axes', 1))
        self.layer.scale_param.bias_term = refine_value(self.op.attributes.get('bias_term', False))


@register_class
class ChannelShuffle(CaffeOpExporter):
    def set_attr(self):
        self.layer.channel_shuffle_param.group = refine_value(self.op.attributes.get('group', 1))


@register_class
class InstanceNormalization(CaffeOpExporter):
    def set_attr(self):
        self.layer.instance_norm_param.num_features = refine_value(self.op.attributes.get('num_features'))
        self.layer.instance_norm_param.eps = refine_value(self.op.attributes.get('eps', 1e-5))
        self.layer.instance_norm_param.affine = refine_value(self.op.attributes.get('affine', False))


@register_class
class Parameter(CaffeOpExporter):
    def set_attr(self):
        self.layer.parameter_param.m = refine_value(self.op.attributes.get('m', -1))
        self.layer.parameter_param.n = refine_value(self.op.attributes.get('n', -1))
        self.layer.parameter_param.batch = refine_value(self.op.attributes.get('batch', 1))
        self.layer.parameter_param.channel = refine_value(self.op.attributes.get('channel', -1))
        self.layer.parameter_param.height = refine_value(self.op.attributes.get('height', -1))
        self.layer.parameter_param.width = refine_value(self.op.attributes.get('width', -1))


@register_class
class Interp(CaffeOpExporter):
    def set_attr(self):
        if refine_value(self.op.attributes.get('shrink_factor')) != 1:
            self.layer.interp_param.shrink_factor = refine_value(self.op.attributes.get('shrink_factor'))
        if refine_value(self.op.attributes.get('zoom_factor')) != 1:
            self.layer.interp_param.zoom_factor = refine_value(self.op.attributes.get('zoom_factor'))
        if refine_value(self.op.attributes.get('width')) and refine_value(self.op.attributes.get('height')):
            self.layer.interp_param.height = refine_value(self.op.attributes.get('height'))
            self.layer.interp_param.width = refine_value(self.op.attributes.get('width'))
        self.layer.interp_param.pad_beg = refine_value(self.op.attributes.get('pad_beg'))
        self.layer.interp_param.pad_end = refine_value(self.op.attributes.get('pad_end'))
        self.layer.interp_param.align_corners = refine_value(self.op.attributes.get('align_corners'))


@register_class
class Tile(CaffeOpExporter):
    def set_attr(self):
        self.layer.tile_param.axis = refine_value(self.op.attributes.get('axis'))
        self.layer.tile_param.tiles = refine_value(self.op.attributes.get('tiles'))


@register_class
class Flatten(CaffeOpExporter):
    def set_attr(self):
        self.layer.flatten_param.axis = refine_value(self.op.attributes.get('axis', 1))
        self.layer.flatten_param.end_axis = refine_value(self.op.attributes.get('end_axis', -1))


@register_class
class SpaceToDepth(CaffeOpExporter):
    def set_attr(self):
        self.layer.subpixel_down_param.downsample = refine_value(self.op.attributes.get('blocksize', 1))


@register_class
class DepthToSpace(CaffeOpExporter):
    def set_attr(self):
        self.layer.subpixel_up_param.upsample = refine_value(self.op.attributes.get('blocksize', 1))


@register_class
class CaffeArgMax(CaffeOpExporter):
    def set_attr(self):
        self.layer.argmax_param.out_max_val = refine_value(self.op.attributes.get('out_max_val'))
        self.layer.argmax_param.top_k = refine_value(self.op.attributes.get('top_k'))
        self.layer.argmax_param.axis = refine_value(self.op.attributes.get('axis'))

@register_class
class HardSwish(CaffeOpExporter):
    pass

@register_class
class HardSigmoid(CaffeOpExporter):
    pass

@register_class
class MatMul(CaffeOpExporter):
    pass




# ---------------------------------------------------------------
# 这个脚本向你展示了如何使用 openvino 对 PPQ 导出的模型进行推理
# 你需要注意，openvino 也可以运行各种各样的量化方案，你甚至可以用 tensorRT 的 policy
# 但总的来说，openvino 需要非对称量化的 activation 和对称量化的 weights
# ---------------------------------------------------------------

# For this onnx inference test, all test data is randomly picked.
# If you want to use real data, just rewrite the defination of SAMPLES

import openvino.runtime
import torch
from ppq import *
from ppq.api import *
from tqdm import tqdm

QUANT_PLATFROM = TargetPlatform.OPENVINO_INT8
MODEL          = 'models\\resnet18.onnx'
INPUT_SHAPE    = [1, 3, 224, 224]
SAMPLES        = [torch.rand(size=INPUT_SHAPE) for _ in range(256)] # rewirte this to use real data.
DEVICE         = 'cuda'
FINETUNE       = False
QS             = QuantizationSettingFactory.default_setting()
EXECUTING_DEVICE = 'cuda'
REQUIRE_ANALYSE  = True

# -------------------------------------------------------------------
# 下面向你展示了常用参数调节选项：
# -------------------------------------------------------------------
QS.lsq_optimization = FINETUNE                                  # 启动网络再训练过程，降低量化误差
QS.lsq_optimization_setting.steps = 500                         # 再训练步数，影响训练时间，500 步大概几分钟
QS.lsq_optimization_setting.collecting_device = 'cuda'          # 缓存数据放在那，cuda 就是放在 gpu，如果显存超了你就换成 'cpu'

# 把量化的不太好的算子送回 FP32
QS.dispatching_table.append(operation='OP NAME', platform=TargetPlatform.FP32)

print('正准备量化你的网络，检查下列设置:')
print(f'TARGET PLATFORM      : {QUANT_PLATFROM.name}')
print(f'NETWORK INPUTSHAPE   : {INPUT_SHAPE}')

# ENABLE CUDA KERNEL 会加速量化效率 3x ~ 10x，但是你如果没有装相应编译环境的话是编译不了的
# 你可以尝试安装编译环境，或者在不启动 CUDA KERNEL 的情况下完成量化：移除 with ENABLE_CUDA_KERNEL(): 即可
with ENABLE_CUDA_KERNEL():
    qir = quantize_onnx_model(
        onnx_import_file=MODEL, calib_dataloader=SAMPLES, calib_steps=128, setting=QS,
        input_shape=INPUT_SHAPE, collate_fn=lambda x: x.to(EXECUTING_DEVICE), 
        platform=QUANT_PLATFROM, do_quantize=True)

    # -------------------------------------------------------------------
    # PPQ 计算量化误差时，使用信噪比的倒数作为指标，即噪声能量 / 信号能量
    # 量化误差 0.1 表示在整体信号中，量化噪声的能量约为 10%
    # 你应当注意，在 graphwise_error_analyse 分析中，我们衡量的是累计误差
    # 网络的最后一层往往都具有较大的累计误差，这些误差是其前面的所有层所共同造成的
    # 你需要使用 layerwise_error_analyse 逐层分析误差的来源
    # -------------------------------------------------------------------
    print('正计算网络量化误差(SNR)，最后一层的误差应小于 0.1 以保证量化精度:')
    reports = graphwise_error_analyse(
        graph=qir, running_device=EXECUTING_DEVICE, steps=32,
        dataloader=SAMPLES, collate_fn=lambda x: x.to(EXECUTING_DEVICE))
    for op, snr in reports.items():
        if snr > 0.1: ppq_warning(f'层 {op} 的累计量化误差显著，请考虑进行优化')

    if REQUIRE_ANALYSE:
        print('正计算逐层量化误差(SNR)，每一层的独立量化误差应小于 0.1 以保证量化精度:')
        layerwise_error_analyse(graph=qir, running_device=EXECUTING_DEVICE,
                                interested_outputs=None,
                                dataloader=SAMPLES, collate_fn=lambda x: x.to(EXECUTING_DEVICE))

    print('网络量化结束，正在生成目标文件:')
    export_ppq_graph(
        graph=qir, platform=QUANT_PLATFROM,
        graph_save_to = 'model_int8.onnx')

    # -------------------------------------------------------------------
    # 启动 openvino 进行推理
    # openvino 跑 int8 挺快的，你的量化策略决定了它能跑多快。
    # 如果你想要了解如何让它跑的更快，你可以在 github 上面联系我。
    # -------------------------------------------------------------------
    openvino_executor = openvino.runtime.Core()
    openvino_results = []
    model = openvino_executor.compile_model(
        model = openvino_executor.read_model(model="model_int8.onnx"), device_name="CPU")
    for sample in tqdm(SAMPLES, desc='OPENVINO GENERATEING OUTPUTS'):
        openvino_results.append(model([convert_any_to_numpy(sample)]))


import torch
from ppq.core import (PPQ_CONFIG, QuantizationProperty, QuantizationStates,
                      RoundingPolicy, TensorQuantizationConfig)
from ppq.utils.round import ppq_tensor_round
from torch.autograd import Function


class TensorwiseLinearQuantImpl(Function):
    """Torch Tensorwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will use ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """
    @ staticmethod
    def forward(ctx, tensor: torch.Tensor, scales: torch.Tensor,
                offsets: torch.Tensor, quant_min: int, quant_max: int,
                rounding: RoundingPolicy) -> torch.Tensor:
        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)

        ctx.save_for_backward(tensor, scales, offsets, torch.tensor(quant_min).type_as(tensor), torch.tensor(quant_max).type_as(tensor), torch.tensor(rounding.value).type(torch.uint8))
        if not PPQ_CONFIG.USING_CUDA_KERNEL or not tensor.is_cuda:
            # quantization function, pytorch implmentation
            tensor = ppq_tensor_round((tensor / scales), rounding) + offsets
            tensor = torch.clamp(tensor, quant_min, quant_max)
            tensor = (tensor - offsets) * scales
            return tensor

        else:
            from ppq.core import CUDA

            # quantization function, pure cuda implmentation
            quantized = CUDA.LinearQuantize_T(
                tensor=tensor,
                scales=scales,
                offsets=offsets,
                minimum=quant_min,
                maximum=quant_max,
                rounding=rounding.value
            )
            return quantized

    @ staticmethod
    def backward(ctx, dy: torch.Tensor):
        x, scales, offsets, qmin, qmax, rounding = ctx.saved_tensors
        rounding = RoundingPolicy(rounding.item())
        tensor = x.clone()
        tensor = tensor / scales
        if rounding == RoundingPolicy.ROUND_HALF_EVEN:
            # default rounding policy of torch is ROUND_TO_NEAR_EVEN
            # try this: print(torch.Tensor([1.5, 2.5, 3.5, 4.5]).round())
            # However it may generate unexpected results due to version difference.
            tensor = tensor.round()
        elif rounding == RoundingPolicy.ROUND_UP:
            tensor = tensor.ceil()
        elif rounding == RoundingPolicy.ROUND_DOWN:
            tensor = tensor.floor()
        elif rounding == RoundingPolicy.ROUND_HALF_TOWARDS_ZERO:
            tensor = torch.sign(tensor) * torch.ceil(tensor.abs() - 0.5)
        elif rounding == RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO:
            tensor = torch.sign(tensor) * torch.floor(tensor.abs() + 0.5)
        elif rounding == RoundingPolicy.ROUND_HALF_DOWN:
            tensor = torch.ceil(tensor - 0.5)
        elif rounding == RoundingPolicy.ROUND_HALF_UP:
            tensor = torch.floor(tensor + 0.5)
        elif rounding == RoundingPolicy.ROUND_TO_NEAR_INT:
            raise NotImplementedError(f'Torch Tensor can not use this rounding policy({rounding}) try ROUND_HALF_EVEN instead.')
        else:
            raise ValueError('Unexpected rounding policy found.')

        tensor = tensor + offsets
        is_lt_min = tensor < qmin
        is_gt_max = tensor > qmax
        is_ge_min_and_le_max = ~is_lt_min & ~is_gt_max

        grad_tensor = dy.clone()
        grad_tensor = torch.where(is_ge_min_and_le_max, grad_tensor, 0 * grad_tensor)

        return grad_tensor.to(dy.device), None, None, None, None, None, None, None, None


class ChannelwiseLinearQuantImpl(Function):
    """Torch Channelwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will bypass this function by using ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """
    @ staticmethod
    def forward(ctx, tensor: torch.Tensor, scales: torch.Tensor,
                offsets: torch.Tensor, channel_axis: int,
                quant_min: int, quant_max: int,
                rounding: RoundingPolicy) -> torch.Tensor:
        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)

        if not PPQ_CONFIG.USING_CUDA_KERNEL or not tensor.is_cuda:
            # generate a shape that likes [1, 1, -1, 1], the only -1 is at channel axe.
            shape = [1 if axis != channel_axis else -1 for axis in range(tensor.ndim)]
            scale, offset = scales.view(shape), offsets.view(shape)

            tensor = ppq_tensor_round((tensor / scale), rounding) + offset
            tensor = torch.clamp(tensor, quant_min, quant_max)
            tensor = (tensor - offset) * scale
            return tensor
        else:
            from ppq.core import CUDA
            quantized = CUDA.LinearQuantize_C(
                tensor=tensor,
                scales=scales,
                offsets=offsets,
                channel_axis=channel_axis,
                minimum=quant_min,
                maximum=quant_max,
                rounding=rounding.value)
            return quantized

    @ staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None, None, None, None, None, None, None, None, None


class TensorwiseDynamicLinearQuantImpl(Function):
    """Torch Tensorwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will use ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """
    @ staticmethod
    def forward(ctx, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        from ppq.quantization.observer.range import minmax_to_scale_offset
        # solve scale and offset at first.
        scales, offsets = minmax_to_scale_offset(
            tensor.min().item(), tensor.max().item(), config=config)
        print(scales, offsets)
        # quantization function, pytorch implmentation
        tensor = ppq_tensor_round((tensor / scales), config.rounding) + offsets
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)
        tensor = (tensor - offsets) * scales
        return tensor

    @ staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None


class ChannelwiseDynamicLinearQuantImpl(Function):
    """Torch Channelwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will bypass this function by using ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """
    @ staticmethod
    def forward(ctx, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        from ppq.quantization.observer.range import minmax_to_scale_offset
        
        channelwise_view = tensor.transpose(dim0=0, dim1=config.channel_axis).unsqueeze(-1)
        channelwise_view = torch.flatten(channelwise_view, start_dim=1)
        
        scales, offsets = [], []
        for _min, _max in zip(
            channelwise_view.min(dim=1)[0].tolist(),
            channelwise_view.max(dim=1)[0].tolist()
        ):
            s, o = minmax_to_scale_offset(_min, _max, config)
            scales.append(s)
            offsets.append(o)

        scales = torch.tensor(scales, dtype=torch.float32, device=tensor.device)
        offsets = torch.tensor(offsets, dtype=torch.float32, device=tensor.device)

        # generate a shape that likes [1, 1, -1, 1], the only -1 is at channel axe.
        shape = [1 if axis != config.channel_axis else -1 for axis in range(tensor.ndim)]
        scales, offsets = scales.view(shape), offsets.view(shape)

        tensor = ppq_tensor_round((tensor / scales), config.rounding) + offsets
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)
        tensor = (tensor - offsets) * scales
        return tensor

    @ staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None

def PPQDyamicLinearQuantFunction(tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    """
    Dynamic Linear Quantization Function(PPQ 动态量化函数).
    
    When calling this method, we firstly solve a scale & offset setting by min-max observer.
    
    Then we applys ordinary Linear Quantization Function with solved setting.
    
    If there is a pre-defined scale & offset within given config, they will be dropped without warning.
    
    动态量化函数将在执行量化之前统计出 tensor 的 min - max, 而后计算出 scale & offset 并完成量化
    
    此时 TQC 中的 scale 与 offset 将被忽略
    """
    if not QuantizationStates.is_activated(config.state): return tensor
    if not config.policy.has_property(QuantizationProperty.LINEAR):
        raise ValueError('Critical Quantization Error! Non-linear config detected.')
    if not config.policy.has_property(QuantizationProperty.DYNAMIC):
        raise ValueError('Quantization Policy Do Not Have Dynamic Flag!')

    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        return ChannelwiseDynamicLinearQuantImpl.apply(tensor, config)
    elif config.policy.has_property(QuantizationProperty.PER_TENSOR):
        return TensorwiseDynamicLinearQuantImpl.apply(tensor, config)

def PPQLinearQuantFunction(
    tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    """PPQ 核心量化函数，没啥好说的了吧，这个玩意既做 quant 也做 dequant"""
    if not QuantizationStates.is_activated(config.state): return tensor
    if not config.policy.has_property(QuantizationProperty.LINEAR):
        raise ValueError('Critical Quantization Error! Non-linear config detected.')
    if config.policy.has_property(QuantizationProperty.DYNAMIC):
        raise ValueError('Unexpected Dynamic Flag in Quantization Policy. Use PPQDyamicQuantFunction Instead.')

    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        return ChannelwiseLinearQuantImpl.apply(
            tensor, config.scale, config.offset, config.channel_axis,
            config.quant_min, config.quant_max, config.rounding)
    elif config.policy.has_property(QuantizationProperty.PER_TENSOR):
        return TensorwiseLinearQuantImpl.apply(
            tensor, config.scale, config.offset,
            config.quant_min, config.quant_max, config.rounding)

def PPQLinearQuant_toInt(tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    """PPQ 核心量化函数，没啥好说的了吧，这个玩意只做 quant 不做 dequant"""
    if not config.policy.has_property(QuantizationProperty.LINEAR):
        raise ValueError('Critical Quantization Error! Non-linear config detected.')

    if config.num_of_bits < 16:
        tensor = tensor.type(dtype=torch.float32)
    elif config.num_of_bits >= 16:
        tensor = tensor.type(dtype=torch.float64)

    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        shape = [1 if axis != config.channel_axis else -1 for axis in range(tensor.ndim)]
        scale, offset = config.scale.view(shape), config.offset.view(shape)
        tensor = ppq_tensor_round((tensor / scale), config.rounding) + offset
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)
    elif config.policy.has_property(QuantizationProperty.PER_TENSOR):
        tensor = ppq_tensor_round((tensor / config.scale), config.rounding) + config.offset
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)

    if config.num_of_bits == 8:
        if config.policy.has_property(QuantizationProperty.SYMMETRICAL):
            return tensor.type(dtype=torch.int8)
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
            return tensor.type(dtype=torch.uint8)
    elif config.num_of_bits == 16:
        if config.policy.has_property(QuantizationProperty.SYMMETRICAL):
            return tensor.type(dtype=torch.int16)
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
            return tensor.type(dtype=torch.uint16)
    elif config.num_of_bits > 16:
        return tensor.type(dtype=torch.int32)
    else: raise Exception('Do not konw how to convert value into int. num of bits is unexpected.')



from typing import Callable, List

import torch
from ppq.core import (NetworkFramework, QuantizationPolicy,
                      QuantizationProperty, RoundingPolicy, TargetPlatform,
                      TensorQuantizationConfig)
from ppq.executor.torch import OPERATION_FORWARD_TABLE
from ppq.IR import BaseGraph, GraphExporter, Variable
from ppq.quantization.observer import BaseTensorObserver, TensorObserverFactroy
from ppq.quantization.optim import (QuantizationOptimizationPass,
                                    QuantizationOptimizationPipeline)
from ppq.quantization.qfunction import PPQuantFunction as QuantFunction
from ppq.quantization.quantizer import BaseQuantizer
from ppq.scheduler import DISPATCHER_TABLE, GraphDispatcher

from .common import __EXPORTERS__, __PARSERS__, __QUANTIZER_COLLECTION__


def Quantizer(platform: TargetPlatform, graph: BaseGraph) -> BaseQuantizer:
    """
    Get a pre-defined Quantizer corresponding to your platform.
    Quantizer in PPQ initializes Tensor Quantization Config for each Operation,
        - it describes how operations are going to be quantized.
    
    根据目标平台获取一个系统预定义的量化器。
    
    ## 量化器
    在 PPQ 中，量化器是一个用于为算子初始化量化信息 Tensor Quantization Config 的对象
        - 量化器决定了你的算子是如何被量化的，你也可以设计新的量化器来适配不同的后端推理框架
    
    在 PPQ 中我们为不同的推理后端设计好了一些预定义的量化器，你可以通过 ppq.lib.Quantizer 来访问它们
    """
    if platform not in __QUANTIZER_COLLECTION__:
        raise KeyError(f'Target Platform {platform} has no related quantizer for now.')
    return __QUANTIZER_COLLECTION__[platform](graph)


def Pipeline(optims: List[QuantizationOptimizationPass]) -> QuantizationOptimizationPipeline:
    """
    Build a Pipeline with given Optimization Passes Collection
    
    使用给定的量化过程集合创建量化管线。
    """
    return QuantizationOptimizationPipeline(optims)


def Observer(
    quant_config: TensorQuantizationConfig, 
    variable: Variable = None) -> BaseTensorObserver:
    """
    Get a Calibration Observer based on quant_config.observer_algorithm attribute.
    
    根据 TQC 中 observer_algorithm 属性获取对应的 Observer.
    """
    return TensorObserverFactroy.build_observer(variable=variable, config=quant_config)


class TensorQuant(torch.nn.Module):
    def __init__(self, quant_config: TensorQuantizationConfig) -> None:
        """
        PPQ Tensor Quant

        Args:
            quant_config (TensorQuantizationConfig): _description_
            name (str, optional): _description_. Defaults to 'PPQ Quant Stub'.
        """
        self._quant_config   = quant_config
        self._delegator      = None
        self._batch_observed = 0
        self._observer       = Observer(quant_config=quant_config)

    @ property
    def delegator(self) -> Callable:
        return self._delegator

    @ delegator.setter
    def delegator(self, func: Callable):
        self._delegator = func

    def forward(self, value: torch.Tensor) -> torch.Tensor:
        if self._delegator is not None:
            return self._delegator(value, self._quant_config)
        return QuantFunction(tensor=value, config=self._quant_config)

    def observe(self, value: torch.Tensor):
        self._batch_observed += 1
        self._observer.observe(value)

    def render(self):
        if self._batch_observed == 0:
            raise PermissionError('You have not provide any data to this QuantStub, '
                                  'PPQ can not render its quant config yet.')
        self._observer.render_quantization_config()


class ParameterQuant(TensorQuant):
    def __init__(self, quant_config: TensorQuantizationConfig, parameter: torch.Tensor) -> None:
        if not isinstance(parameter, torch.Tensor):
            raise TypeError(f'Expect a torch.Tensor here. However {type(parameter)} was given.')
        
        super().__init__(quant_config)
        self.observe(parameter)
        self.render()


def LinearQuantizationConfig(
    symmetrical: bool = True,
    dynamic: bool = False,
    power_of_2: bool = False,
    channel_axis: int = None,
    quant_min: int = -128,
    quant_max: int = 127,
    num_of_bits = 8,
    calibration: str = 'minmax',
    rounding: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN) -> TensorQuantizationConfig:
    """
    Create a Linear Quantization Config.
    
    创建线性量化配置信息。
    """

    sym = QuantizationProperty.SYMMETRICAL if symmetrical else QuantizationProperty.ASYMMETRICAL
    dyn = QuantizationProperty.DYNAMIC if dynamic else 0
    pw2 = QuantizationProperty.POWER_OF_2 if power_of_2 else 0
    chn = QuantizationProperty.PER_TENSOR if channel_axis is None else QuantizationProperty.PER_CHANNEL

    return TensorQuantizationConfig(
        policy = QuantizationPolicy(sym + dyn + pw2 + chn + QuantizationProperty.LINEAR),
        rounding = rounding,
        num_of_bits = num_of_bits,
        quant_min = quant_min,
        quant_max = quant_max,
        observer_algorithm = calibration,
        channel_axis=channel_axis)


def FloatingQuantizationConfig(
    symmetrical: bool = True,
    power_of_2: bool = True,
    channel_axis: int = None,
    quant_min: float = -448.0,
    quant_max: float = 448.0,
    exponent: int = 4,
    mantissa: int = 3,
    calibration: str = 'constant',
    rounding: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN) -> TensorQuantizationConfig:
    """
    Create a Floating Quantization Config.
    
    创建浮点量化配置信息。
    """

    sym = QuantizationProperty.SYMMETRICAL if symmetrical else QuantizationProperty.ASYMMETRICAL
    pw2 = QuantizationProperty.POWER_OF_2 if power_of_2 else 0
    chn = QuantizationProperty.PER_TENSOR if channel_axis is None else QuantizationProperty.PER_CHANNEL

    return TensorQuantizationConfig(
        policy = QuantizationPolicy(sym + pw2 + chn + QuantizationProperty.FLOATING),
        rounding = rounding,
        num_of_bits = exponent + mantissa + 1,
        exponent_bits = exponent,
        quant_min = quant_min,
        quant_max = quant_max,
        observer_algorithm = calibration)


def Dispatcher(graph: BaseGraph, method: str='conservative') -> GraphDispatcher:
    """
    Get a Graph Dispatcher.
    
    获取一个指定的调度器。
    """
    if method not in DISPATCHER_TABLE:
        raise KeyError(f'Can not find a dispatcher named {method}, check your input again.')
    return DISPATCHER_TABLE[method](graph)


def OperationForwardFunction(optype: str, platform: TargetPlatform = TargetPlatform.FP32) -> Callable:
    """
    Get an Operation forward function. Same Op are allows to have different forward function on different platform,
    to get a default forward function, use platform=TargetPlatform.FP32.
    
    获取一个算子前向传播执行函数。在 PPQ 中，相同的算子可以在不同的平台上注册成不同的执行逻辑，
    使用 platform = TargetPlatform.FP32 来获取默认执行逻辑。
    """
    if not isinstance(platform, TargetPlatform):
        raise TypeError('Wrong parameter type for invoking this function.')
    if optype not in OPERATION_FORWARD_TABLE[platform]:
        raise KeyError(f'Can not find a forward function related to optype {optype}({platform.name}),'
                       ' Register it first.')
    return OPERATION_FORWARD_TABLE[platform][optype]


def Exporter(platform: TargetPlatform) -> GraphExporter:
    """
    Get an network Exporter.
    
    获取一个网络导出器。
    """
    if not isinstance(platform, TargetPlatform):
        raise TypeError('Wrong parameter type for invoking this function.')
    if platform not in __EXPORTERS__:
        raise KeyError(f'Platfrom {platform.name} has no related exporter, register a exporter for it first.')
    return __EXPORTERS__[platform]()


def Parser(framework: NetworkFramework) -> GraphExporter:
    """
    Get an network Parser.
    
    获取一个网络解析器。
    """
    if not isinstance(framework, NetworkFramework):
        raise TypeError('Parameter framework has invalid type, Check your parameter again.')
    if framework not in __PARSERS__:
        raise KeyError(f'Requiring framework {framework} does not support parsing now.')
    return __PARSERS__[framework]()


__all__ = [
    'Parser', 'Exporter', 'OperationForwardFunction', 
    'Dispatcher', 'FloatingQuantizationConfig', 'LinearQuantizationConfig', 
    'QuantStub', 'Quantizer', 'Observer', 'Pipeline', 'QuantFunction']
